# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from tamkin.data import Molecule, RotScan

from molmod.io import FCHKFile
from molmod import dihed_angle, amu, angstrom

import numpy as np


__all__ = [
    "load_fixed_g03com", "load_molecule_g03fchk", "load_molecule_g98fchk",
    "load_rotscan_g03log",
]


def load_fixed_g03com(filename):
    """Load fixed atoms from a gaussian input file.

       Argument:
        | ``filename`` -- The gaussian input file.

       A fixed atom is recognized by the '-1' after the atom symbol in the
       molecule specification. The '-1' is the second word in the line.
    """
    f = file(filename)
    for line in f:
        # iterate until we reach the title line
        line = line.strip()
        if not (len(line)==0 or line[0] == "#" or line[0] == "%"):
            break
    for line in f:
        # skip lines until charge/multiplicty is reached
        words = line.split()
        if len(words) == 2:
            try:
                int(words[0])
                int(words[1])
                break
            except ValueError:
                pass
    counter = 0
    fixed_atoms = []
    for line in f:
        # go trough the list of atoms and store the fixed ones.
        words = line.split()
        if len(words) == 0:
            break
        if words[1] == "-1":
            fixed_atoms.append(counter)
        counter += 1
    return fixed_atoms


def iter_floats_file(fn):
    f = file(fn)
    for line in f:
        words = line.split()
        for w in words:
            yield float(w.replace('D', 'E'))
    f.close()


def load_molecule_g03fchk(fn_freq, fn_ener=None, fn_vdw=None, energy=None, fn_punch=None):
    """Load a molecule from Gaussian03 formatted checkpoint files.

       Arguments:
         | ``fn_freq`` -- the formatted checkpoint file of the frequency job

       Optional arguments:
         | ``fn_ener`` -- the formatted checkpoint file of a single point
                          computation for the energy. When not given, the energy
                          is taken from the frequency job.
         | ``fn_vdw`` -- An orca output file containing a Van der Waals
                         correction for the energy.
         | ``energy`` -- Override the energy from the formatted checkpoint file
                         with the given value.
         | ``punch`` -- A Gaussian derivatives punch file. When given, the
                        gradient and the Hessian are read from this file
                        instead.
    """

    fchk_freq = FCHKFile(fn_freq, ignore_errors=True, field_labels=[
        "Cartesian Force Constants", "Real atomic weights", "Total Energy",
        "Multiplicity", "Cartesian Gradient", "MicOpt",
    ])
    if fn_ener is None:
        fchk_ener = fchk_freq
    else:
        fchk_ener = FCHKFile(fn_ener, ignore_errors=True, field_labels=[
            "Total Energy"
        ])
    if energy is None:
        energy = fchk_ener.fields["Total Energy"]
    vdw = 0
    if fn_vdw is not None:
        from tamkin.io.dispersion import load_dftd_orca
        vdw = load_dftd_orca(fn_vdw)

    natom = fchk_freq.molecule.size
    if fchk_freq.molecule.size == 1 and \
       "Cartesian Force Constants" not in fchk_freq.fields:
        gradient = np.zeros((1,3), float)
        hessian = np.zeros((3,3), float)
    elif fn_punch is None:
        gradient = fchk_freq.fields["Cartesian Gradient"].copy()
        gradient.shape = (natom, 3)
        hessian = fchk_freq.get_hessian()
    else:
        gradient = np.zeros((natom, 3), float)
        hessian = np.zeros((3*natom, 3*natom), float)
        iterator = iter_floats_file(fn_punch)
        for i in xrange(natom):
            for j in xrange(3):
                gradient[i,j] = iterator.next()
        for i in xrange(3*natom):
            for j in xrange(i+1):
                v = iterator.next()
                hessian[i,j] = v
                hessian[j,i] = v

    if "MicOpt" in fchk_freq.fields:
        fixed = (fchk_freq.fields["MicOpt"] == -2).nonzero()[0]
        if len(fixed) == 0:
            fixed = None
    else:
        fixed = None

    return Molecule(
        fchk_freq.molecule.numbers,
        fchk_freq.molecule.coordinates,
        fchk_freq.fields["Real atomic weights"]*amu,
        energy+vdw,
        gradient,
        hessian,
        fchk_freq.fields["Multiplicity"],
        None, # gaussian is very poor at computing the rotational symmetry number
        False,
        title=fchk_freq.title,
        fixed=fixed,
    )


g98_masses = np.array([
    1.0079, 4.0026, 6.94, 9.01218, 10.81, 12.011, 14.0067, 15.9994, 18.9984,
    20.179, 22.98977, 24.305, 26.98154, 28.0855, 30.97376, 32.06, 35.453,
    39.948, 39.0983, 40.08, 44.9559, 47.9, 50.9415, 51.996, 54.938, 55.847,
    58.9332, 58.71, 63.546, 65.38, 69.735, 72.59, 74.9216, 78.96, 79.904, 83.8,
    85.4678, 87.62, 88.9059, 91.22, 92.9064, 95.94, 98.9062, 101.07, 102.9055,
    106.4, 107.868, 112.41, 114.82, 118.69, 121.75, 127.6, 126.9045, 131.3,
    132.9054, 137.33, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
    0.0, 0.0, 0.0, 0.0, 178.49, 180.9479, 183.85, 186.207, 190.2, 192.22,
    195.09, 196.9665, 200.59, 204.37, 207.2, 208.9804, 0.0, 0.0, 0.0, 0.0, 0.0,
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
    0.0, 0.0, 0.0, 0.0,
])*amu


def load_molecule_g98fchk(fn_freq, fn_ener=None, energy=None):
    """Load a molecule from Gaussian98 formatted checkpoint files.

       Arguments:
         | ``fn_freq`` -- The formatted checkpoint file of the frequency job.

       Optional arguments:
         | ``fn_ener`` -- The formatted checkpoint file of a single point
                          computation for the energy. When not given, the energy
                          is taken from the frequency job.
         | ``energy`` -- Override the energy from the formatted checkpoint file
                         with the given value.
    """

    fchk_freq = FCHKFile(fn_freq, ignore_errors=True, field_labels=[
        "Cartesian Force Constants", "Total Energy",
        "Multiplicity", "Cartesian Gradient"
    ])
    if fn_ener is None:
        fchk_ener = fchk_freq
    else:
        fchk_ener = FCHKFile(fn_ener, ignore_errors=True, field_labels=[
            "Total Energy"
        ])
    masses = np.array([g98_masses[n-1] for n in fchk_freq.molecule.numbers])
    if energy is None:
        energy = fchk_ener.fields["Total Energy"]

    return Molecule(
        fchk_freq.molecule.numbers,
        fchk_freq.molecule.coordinates,
        masses,
        energy,
        np.reshape(np.array(fchk_freq.fields["Cartesian Gradient"]), (len(fchk_freq.molecule.numbers),3)),
        fchk_freq.get_hessian(),
        fchk_freq.fields["Multiplicity"],
        None, # gaussian is very poor at computing the rotational symmetry number
        False,
    )


def load_rotscan_g03log(fn_log, top_indexes=None):
    """Load the torsional potential from a Gaussian 03 log/output file.

       Argument:
        | ``fn_log`` -- The filename of the gaussian output.

       Optional argument:
        | ``top_indexes`` -- The atom indexes that define the rotor. These do
                             not have to include the atoms that define the
                             rotational axis. When not given, an attempt is made
                             to derive this information from the dihedral angle
                             used for the scan.
    """
    # find the line that specifies the dihedral angle
    with open(fn_log) as f:
        found_modredundant = False
        for line in f:
            if line.startswith(" The following ModRedundant input section has been read:"):
                found_modredundant = True
                break
        if not found_modredundant:
            raise IOError("Could not find the ModRedundant section in the log file.")

        dihedral = None
        num_geoms = 0
        for line in f:
            line = line.strip()
            if len(line) == 0:
                break
            else:
                words = line.split()
                if words[0] == "D" and (words[5] == "S" or words[5] == "F"):
                    if dihedral is None:
                        dihedral = list(int(word)-1 for word in words[1:5])
                    else:
                        raise IOError("Found multiple dihedral angle scan, which is not supported.")

        if dihedral is None:
            raise IOError("Could not find the dihedral angle of the rotational scan.")

        # load all the energies and compute the corresponding angles
        energies = []
        angles = []
        geometries = []
        for line in f:
            if line.startswith("                          Input orientation:") or \
               line.startswith("                         Standard orientation:"):
                # read the molecule
                numbers = []
                last_coordinates = []
                ## skip four lines
                for i in xrange(4):
                    f.next()
                # read atoms
                for line in f:
                    if line.startswith(" -----"):
                        break
                    words = line.split()
                    numbers.append(int(words[1]))
                    last_coordinates.append((float(words[3]), float(words[4]), float(words[5])))
            if line.startswith(" SCF Done:"):
                # read the energy
                last_energy = float(line[line.find("=")+1:].split()[0])
            if line.startswith("    -- Stationary point found."):
                # store last emergy and geometry in list
                energies.append(last_energy)
                last_coordinates = np.array(last_coordinates)*angstrom
                geometries.append(last_coordinates)
                angles.append(dihed_angle(last_coordinates[dihedral])[0])

        if len(energies) == 0:
            raise IOError("Could not find any stationary point")

        if top_indexes is None:
            # Define the molecular geometry that is used in the constructor of
            # RotScan to detect the top.
            from molmod.molecules import Molecule as BaseMolecule
            molecule = BaseMolecule(numbers, geometries[0])
        else:
            molecule = None
        result = RotScan(
            dihedral, molecule, top_indexes,
            np.array([angles, energies])
        )
        result.geometries = np.array(geometries)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


import numpy as np

from tamkin import *
from molmod import *


# define some parameters
fixed = [6,7,8,9,10,11]
level = "m062x"
temp = 87.0*kelvin

# load molecules
mol_a = load_molecule_g03fchk("%s/argon/gaussian.fchk" % level)
mol_b = load_molecule_g03fchk("%s/benzene/freq/gaussian.fchk" % level)
mol_c = load_molecule_g03fchk("%s/complex/freq/gaussian.fchk" % level)

# do the normal mode analysis
nma_a = NMA(mol_a, ConstrainExt()) # does not much, but is mandatory due to API
nma_b = NMA(mol_b, PHVA(fixed)) # use PHVA to model the surface
nma_c = NMA(mol_c, PHVA(fixed)) # use PHVA to model the surface

# Add a small correction to the energy from the m062x computation as to get a
# perfect correspondence between the experimental and theoretical isotherm in
# the ideal gas limit There are several things that can cause errors in the
# m062x computation: poor graphite model (i.e. benzene), lack of BSSE
# correction, imperfections in the DFT functional. The first and the second
# error will probably compensate partially. The benzene approximation is rather
# dramatic because the Van der Waals sphere of Argon is larger than benzene.
nma_c.energy += -2.4*kjmol
# Note that kT at 87K is about 0.723 kJ/mol. This is example a very sensitive
# to the accuracy of the adsorption energy.

# setup an NpT partition function for 3D gas
pf_a = PartFun(nma_a, [ExtTrans(cp=True)])
# benzene acts a a substrate fixed in space, no translation or rotation
pf_b = PartFun(nma_b, [])
# setup an NVT partition function for the 2D gas. (The amount of surface is
# constant.) Specify that atom with index 12 (Argon) is the mobile one.
pf_c = PartFun(nma_c, [ExtTrans(cp=False, mobile=[12], dim=2)])
# create a model for the thermodynamic equilibrium
tm = ThermodynamicModel([pf_a, pf_b], [pf_c])
tm.write_to_file("adsorption_%s.txt" % level)
# reference data from http://dx.doi.org/10.1021/jp011745+, table 4
#   The argon cross section in the paper is 0.138 nm**2
#   This example is based on data at temperature of 87 K.
# first column: p/p0
#   with p0 = 0.96848*atm
# second column: alpha_s
#   Alpha_s is the ratio of the adsorbed amount over a reference, i.e. 3.108
#   cm**3 STP g**-1. The adsorbent has a specific surface area of 6.01 m**2
#   g**-1 and can be treated as a macroscopic graphite surface. We model this
#   surface as a graphene layer, which is further approximated by a single
#   benzene molecule. ouch.
ref_data = np.array([
    (1.28e-6, 9.17e-5),
    (1.28e-4, 9.17e-3),
    (2.49e-4, 0.0184),
    (3.54e-4, 0.0274),
    (4.50e-4, 0.0364),
    (5.38e-4, 0.0456),
    (6.20e-4, 0.0548),
    (6.95e-4, 0.0641),
    (7.65e-4, 0.0735),
    (8.30e-4, 0.0829),
    (8.91e-4, 0.0924),
    (9.49e-4, 0.102),
    (1.00e-3, 0.111),
    (1.06e-3, 0.121),
    (1.11e-3, 0.131),
    (1.16e-3, 0.14),
    (1.20e-3, 0.15),
    (1.25e-3, 0.16),
    (1.30e-3, 0.169),
    (1.34e-3, 0.179),
    (1.39e-3, 0.188),
    (1.43e-3, 0.198),
    (1.48e-3, 0.208),
    (1.52e-3, 0.217),
    (1.57e-3, 0.227),
    (1.62e-3, 0.237),
    (1.67e-3, 0.246),
    (1.72e-3, 0.256),
    (1.77e-3, 0.265),
    (1.83e-3, 0.275),
    (1.89e-3, 0.284),
    (1.96e-3, 0.294),
    (2.03e-3, 0.303),
    (2.10e-3, 0.312),
    (2.18e-3, 0.322),
    (2.27e-3, 0.331),
    (2.36e-3, 0.34),
    (2.47e-3, 0.349),
    (2.58e-3, 0.358),
    (2.71e-3, 0.366),
    (2.86e-3, 0.375),
    (3.02e-3, 0.383),
    (3.19e-3, 0.391),
    (3.39e-3, 0.399),
    (3.62e-3, 0.406),
    (3.86e-3, 0.414),
    (4.13e-3, 0.42),
    (4.44e-3, 0.427),
    (4.81e-3, 0.434),
    (5.23e-3, 0.441),
    (5.69e-3, 0.447),
    (6.20e-3, 0.452),
    (6.75e-3, 0.458),
    (7.55e-3, 0.465),
    (8.13e-3, 0.469),
    (8.75e-3, 0.473),
    (9.40e-3, 0.477),
    (0.01, 0.48),
    (0.0113, 0.486),
    (0.0125, 0.491),
    (0.015, 0.498),
    (0.0175, 0.505),
    (0.0202, 0.511),
    (0.025, 0.518),
    (0.0301, 0.525),
    (0.0351, 0.531),
    (0.0401, 0.536),
    (0.0452, 0.541),
    (0.0501, 0.545),
    (0.0573, 0.551),
    (0.0695, 0.56),
    (0.0793, 0.567),
    (0.089, 0.573),
    (0.0987, 0.58),
    (0.117, 0.591),
    (0.137, 0.602),
    (0.156, 0.616),
    (0.176, 0.633),
    (0.195, 0.649),
    (0.215, 0.671),
    (0.235, 0.694),
    (0.256, 0.719),
    (0.276, 0.746),
    (0.297, 0.778),
    (0.317, 0.817),
    (0.337, 0.861),
    (0.36, 0.914),
    (0.381, 0.959),
    (0.4, 1.0),
    (0.42, 1.04),
    (0.44, 1.07),
    (0.46, 1.1),
    (0.48, 1.13),
    (0.5, 1.15),
    (0.52, 1.18),
    (0.54, 1.21),
    (0.56, 1.23),
    (0.58, 1.26),
    (0.6, 1.3),
    (0.62, 1.34),
    (0.64, 1.38),
    (0.66, 1.44),
    (0.68, 1.5),
    (0.7, 1.57),
    (0.72, 1.65),
    (0.74, 1.72),
    (0.76, 1.78),
    (0.78, 1.84),
    (0.8, 1.91),
    (0.819, 1.98),
    (0.84, 2.07),
    (0.86, 2.17),
    (0.879, 2.3),
    (0.899, 2.45),
    (0.919, 2.63),
    (0.94, 2.9),
    (0.954, 3.27),
    (0.964, 3.63),
    (0.975, 4.22),
    (0.984, 5.09),
    (0.988, 5.9),
])[:63]

# convert first column to pressure
p0 = 0.981313*bar
exp_pressure = ref_data[:,0]*p0
# convert second column to occupation, i.e. the number of argon molecules per
# site. A site is defined as a part of the area that has the size of the Argon
# cross-section.
# STD stands for standard temperature and pressure. IUPAC says:
#    standard pressure = 1 bar
#    standard temperature = 273.15 K
alpha_ref = 3.108*centimeter**3/gram # from paper
stp_dens_ar = 44.072*mol/meter**3    # density of argon under STP conditions
spec_area = 6.01*meter**2/gram       # specific area of the graphite sample in the paper
cross_sec_ar = 0.138*nanometer**2    # the cross section of one argon site on the surface
occ_reference = alpha_ref*stp_dens_ar/spec_area*cross_sec_ar
exp_occupation = ref_data[:,1]*occ_reference

# compute the ideal gas estimate of the occupation using the thermodynamic model
mod_occupation = []
for p in exp_pressure:
    # compute the density of argon in the gas phase (3D)
    dens_ar = p/(boltzmann*temp)
    # use the equilibrium constant of the thermodynamic model to compute the
    # density on the 2D surface.
    K = tm.equilibrium_constant(temp)
    dens2d = K*dens_ar
    # from density to occupation using the cross section
    occupation = dens2d*cross_sec_ar
    mod_occupation.append(occupation)
mod_occupation = np.array(mod_occupation)

# Make a plot of the adsorption isotherm with the ideal gas and with the
# Langmuir model. Remaining deviations with experiment are due to (relatively
# strong) Ar-Ar interactions on the graphite surface, i.e. the 2D gas is far
# from ideal. One must do Monte-Carlo simulations to estimate these effects.
import matplotlib.pyplot as pt
pt.clf()
pt.plot(exp_pressure/bar, exp_occupation, 'k-', lw=2, label='experiment')
pt.plot(exp_pressure/bar, mod_occupation, 'b-', lw=1, label='TAMkin ideal gas')
pt.plot(exp_pressure/bar, mod_occupation/(1+mod_occupation), 'g-', lw=1, label='TAMkin Langmuir')
pt.ylim(0,1.0)
pt.xlim(0,5e-3)
pt.ylabel("Fraction of occupied mono-layer\nArgon sites on graphite at T=87K")
pt.xlabel("Pressure [bar]")
pt.legend(loc=0)
pt.savefig("adsorption_%s_isotherm.png" % level)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from lot_basis import lots_list
import html
from kin import *

from molmod.units import kjmol
from molmod.constants import boltzmann

import numpy as np



def overview(template, title, fn_img, rows):
    if len(rows) == 0:
        rows.append([u"<th>Approach→<br />Functional↓</th>"])
        rows.append(["<th></th>"])
    rows[0].append("<th colspan=4>%s</th>" % title.replace(", ", "<br />"))
    for temp in temps:
        rows[1].append("<td style='font-size:small'>%.0fK</td>" % temp)
    counter = 2
    for lot in lots_list:
        if lot.spin == "ROS":
            lot_label = "ro" + lot.label
        else:
            lot_label = lot.label
        if len(rows) <= counter:
            rows.append(["<th>%s</th>" % lot_label])
        try:
            ks = load_summary(template % lot_label)[0]
            for j in xrange(4):
                ln10ratio = np.log10(ks[j]/experimental_k[j])
                color = get_error_color(ln10ratio)
                rows[counter].append("<td style='background-color:%s'>%.1e</td>" % (color, ks[j]))
        except IOError, StopIteration:
            rows[counter].append("<td>&nbsp</td><td>&nbsp</td><td>&nbsp</td><td>&nbsp</td>")
        counter += 1




f = file("abskintab.html", "w")
print >> f, html.header % "KIN Overview"

for do_rotor in False, True:
    ir_str = {True: "ir", False: "ho"}[do_rotor]
    ir_info = {
        True: "internal rotor",
        False: "harmonic oscillator",
    }[do_rotor]
    for do_counterpoise in False, True:
        cp_str = {True: "cps", False: "bss"}[do_counterpoise]
        cp_info = {
            True: "with counterpoise correction",
            False: "without counterpoise correction",
        }[do_counterpoise]

        rows = []

        for ts_conformer in "Gauche", "Trans":
            overview(
                "%%s__6-31gd/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, Consistent, 6-31G(d)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_consistent_6-31gd_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )
        for ts_conformer in "Gauche", "Trans":
            overview(
                "%%s__6-311+g3df2p/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, Consistent, 6-311+G(3df,2p)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_consistent_6-311+g3df2p_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )
        for ts_conformer in "Gauche", "Trans":
            overview(
                "GEO__b3lyp__6-31gd__ENERGY__%%s__6-311+g3df2p/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, GEO=B3LYP/6-31G(d), 6-311+G(3df,2p)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_geo_6-311+g3df2p_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )

        print >> f, "<p>Theoretical rate constants (%s, %s)</p>" % (ir_info, cp_info)
        html.print_table(f, rows)

print >> f, html.footer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from molmod.io.psf import PSFFile
from molmod.io.xyz import XYZFile
from molmod.molecular_graphs import MolecularGraph, HasAtomNumber, HasNeighbors, HasNeighborNumbers
from molmod.graphs import CritAnd


import sys


CC30A = CritAnd(HasAtomNumber(6), HasNeighborNumbers(6,6,6,6))
CC31A = CritAnd(HasAtomNumber(6), HasNeighborNumbers(6,6,6,1))
CC32A = CritAnd(HasAtomNumber(6), HasNeighborNumbers(6,6,1,1))
CC33A = CritAnd(HasAtomNumber(6), HasNeighborNumbers(6,1,1,1))

atom_filters = {
    "CC30A": CC30A,
    "CC31A": CC31A,
    "CC32A": CC32A,
    "CC33A": CC33A,
    "HCA1": CritAnd(HasAtomNumber(1), HasNeighbors(CC31A)),
    "HCA2": CritAnd(HasAtomNumber(1), HasNeighbors(CC32A)),
    "HCA3": CritAnd(HasAtomNumber(1), HasNeighbors(CC33A)),
}

def get_atom_type(index, graph):
    for atom_type, atom_filter in atom_filters.iteritems():
        if atom_filter(index, graph):
            return atom_type
    raise ValueError("Unrecognized atom (index %i)." % index)

args = sys.argv[1:]

molecule = XYZFile(args[0]).get_molecule()
graph = MolecularGraph.from_geometry(molecule)
atom_types = [get_atom_type(index, graph) for index in xrange(molecule.size)]

psf = PSFFile()
psf.add_molecular_graph(graph, atom_types=atom_types)
psf.write_to_file(args[0].replace(".xyz", ".psf"))



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Timer to keep track of wall and cpu time"""


import time, sys


__all__ = ["Timer"]


class Timer(object):
    """Timer object to keep track of the time spent in several parts of TAMKin.

       This timer works like a stopwatch. Each time the :meth:`sample` method is
       called the current cpu and wall times are recorded toghether with a label

       The methods :meth:`dump` and :meth:`write_to_file` can be used to
       generate a report.
    """

    def __init__(self):
        self.cpu_times = []
        self.wall_times = []
        self.labels = []

    def sample(self, label):
        """Record the current timings and associate them with the given label

           Argument:
            | label  --  A string describing this point in the code.
        """
        self.cpu_times.append(time.clock())
        self.wall_times.append(time.time())
        self.labels.append(label)

    def dump(self, f=sys.stdout):
        """Dump the logfile with timing information, to screen or to a file stream.

           Optional argument:
            | f  --  the stream to write to. [default=sys.stdout]
        """
        print >> f, "-------------------"
        print >> f, "Printing LOG jobtimer"
        print >> f, '%12s %12s %21s %16s %30s' %("cpu times [s]", "diff [s]", "wall times [s]", "diff [s]", "labels" )
        for i,label in enumerate(self.labels[:-1]):
            print >> f, '%12.3f %12.3f %21.3f %16.3f %30s' %(self.cpu_times[i],
                                         self.cpu_times[i+1]-self.cpu_times[i],
                                         self.wall_times[i],
                                         self.wall_times[i+1]-self.wall_times[i],
                                         label)
        print >> f, '%12.3f %12s %21.3f %16s %30s' %(self.cpu_times[-1], "",
                                         self.wall_times[-1], "",
                                         self.labels[-1])
        print >> f, "-------------------"

    def write_to_file(self, filename):
        """Write the logfile with timing information to filename.

           Argument:
            | filename  --  the file to write to.
        """
        f = file(filename, 'w')
        self.dump(f)
        f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin library.
from tamkin import *
# Import units
from molmod.units import kjmol
# import matplotlib.pyplot as pt for plotting
import matplotlib.pyplot as pt
# Import standard python libraries
import os, sys


def load_rotor(mol, filename, rotsym, even, expansion=5):
    rot_scan = load_rotscan_g03log(filename)
    rotor = Rotor(rot_scan, mol, rotsym=rotsym, even=even)
    return rotor


def load_cps_barrier(fn_template):
    """Loads the counterpoise corrected barrier from five sp calculations"""
    from molmod.io import FCHKFile
    def load_ener(fn_fchk):
        fchk = FCHKFile(fn_fchk, field_labels=["Total Energy"])
        return fchk.fields["Total Energy"]
    return (
        load_ener(fn_template % "full_0") +
        load_ener(fn_template % "sole_1") - load_ener(fn_template % "full_1") +
        load_ener(fn_template % "sole_2") - load_ener(fn_template % "full_2")
    )



def run(do_rotor, do_counterpoise, load_sp):
    prefix = {True: "ir", False: "ho"}[do_rotor]
    prefix += {True: "_cps", False: "_bss"}[do_counterpoise]

    if load_sp:
        mol_ethyl = load_molecule_g03fchk("ethyl__freq/gaussian.fchk", "ethyl__sp/gaussian.fchk")
        mol_ethene = load_molecule_g03fchk("ethene__freq/gaussian.fchk", "ethene__sp/gaussian.fchk")
    else:
        mol_ethyl = load_molecule_g03fchk("ethyl__freq/gaussian.fchk")
        mol_ethene = load_molecule_g03fchk("ethene__freq/gaussian.fchk")
    if do_counterpoise:
        try:
            mol_ts_gauche = load_molecule_g03fchk("ts_ad1_gauche__freq/gaussian.fchk", "ts_ad1_gauche__bsse/gaussian.fchk")
        except (IOError, KeyError):
            cps_barrier_gauche = load_cps_barrier("ts_ad1_gauche__cps_%s/gaussian.fchk")
            mol_ts_gauche = load_molecule_g03fchk("ts_ad1_gauche__freq/gaussian.fchk", energy=cps_barrier_gauche)
        try:
            mol_ts_trans = load_molecule_g03fchk("ts_ad1_trans__freq/gaussian.fchk", "ts_ad1_trans__bsse/gaussian.fchk")
        except (IOError, KeyError):
            cps_barrier_trans = load_cps_barrier("ts_ad1_trans__cps_%s/gaussian.fchk")
            mol_ts_trans = load_molecule_g03fchk("ts_ad1_trans__freq/gaussian.fchk", energy=cps_barrier_trans)
    else:
        if load_sp:
            mol_ts_gauche = load_molecule_g03fchk("ts_ad1_gauche__freq/gaussian.fchk", "ts_ad1_gauche__sp/gaussian.fchk")
            mol_ts_trans = load_molecule_g03fchk("ts_ad1_trans__freq/gaussian.fchk", "ts_ad1_gauche__sp/gaussian.fchk")
        else:
            mol_ts_gauche = load_molecule_g03fchk("ts_ad1_gauche__freq/gaussian.fchk")
            mol_ts_trans = load_molecule_g03fchk("ts_ad1_trans__freq/gaussian.fchk")
    # Perform normal mode analysis on the molecules
    nma_ethyl = NMA(mol_ethyl, ConstrainExt(gradient_threshold=1e-3))
    nma_ethene = NMA(mol_ethene, ConstrainExt(gradient_threshold=1e-3))
    nma_ts_gauche = NMA(mol_ts_gauche, ConstrainExt(gradient_threshold=1e-3))
    nma_ts_trans = NMA(mol_ts_trans, ConstrainExt(gradient_threshold=1e-3))
    if do_rotor:
        # Construct the rotors
        rotor_ethyl = load_rotor(mol_ethyl, "ethyl__scan_methyl/gaussian.log", 6, True, 1)
        rotor1_ts_gauche = load_rotor(mol_ts_gauche, "ts_ad1_gauche__scan_methyl/gaussian.log", 3, False)
        rotor2_ts_gauche = load_rotor(mol_ts_gauche, "ts_ad1_trans__scan_forming_bond/gaussian.log", 1, True)
        rotor1_ts_trans = load_rotor(mol_ts_trans, "ts_ad1_trans__scan_methyl/gaussian.log", 3, False)
        rotor2_ts_trans = load_rotor(mol_ts_trans, "ts_ad1_trans__scan_forming_bond/gaussian.log", 1, True)
        # Construct the partition functions.
        pf_ethyl = PartFun(nma_ethyl, [
            ExtTrans(), ExtRot(), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
            rotor_ethyl,
        ])
        pf_ethene = PartFun(nma_ethene, [
            ExtTrans(), ExtRot(),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
        ])
        pf_ts_gauche = PartFun(nma_ts_gauche, [
            ExtTrans(), ExtRot(), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
            rotor1_ts_gauche, rotor2_ts_gauche,
        ])
        pf_ts_trans = PartFun(nma_ts_trans, [
            ExtTrans(), ExtRot(), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
            rotor1_ts_trans, rotor2_ts_trans,
        ])
    else:
        # Construct the partition functions.
        pf_ethyl = PartFun(nma_ethyl, [
            ExtTrans(), ExtRot(1), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
        ])
        pf_ethene = PartFun(nma_ethene, [
            ExtTrans(), ExtRot(4),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
        ])
        pf_ts_gauche = PartFun(nma_ts_gauche, [
            ExtTrans(), ExtRot(1), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
        ])
        pf_ts_trans = PartFun(nma_ts_trans, [
            ExtTrans(), ExtRot(1), Electronic(2),
            #Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
        ])

    if do_rotor:
        # Plot the energy levels and the potential of the hindered rotor. The
        # temperature argument is used to indicate the population of each level in the
        # plot.
        rotor_ethyl.plot_levels("rotor_ethyl_energy_levels.png", 300)
        rotor1_ts_gauche.plot_levels("rotor1_ts_gauche_energy_levels.png", 300)
        rotor2_ts_gauche.plot_levels("rotor2_ts_gauche_energy_levels.png", 300)
        rotor1_ts_trans.plot_levels("rotor1_ts_trans_energy_levels.png", 300)
        rotor2_ts_trans.plot_levels("rotor2_ts_trans_energy_levels.png", 300)


    # Analyse the partition functions in detail
    ta_ethyl = ThermoAnalysis(pf_ethyl, [300,400,500,600])
    ta_ethyl.write_to_file("%s_thermo_ethyl.csv" % prefix)
    ta_ethene = ThermoAnalysis(pf_ethene, [300,400,500,600])
    ta_ethene.write_to_file("%s_thermo_ethene.csv" % prefix)
    ta_ts_gauche = ThermoAnalysis(pf_ts_gauche, [300,400,500,600])
    ta_ts_gauche.write_to_file("%s_thermo_ts_gauche.csv" % prefix)
    ta_ts_trans = ThermoAnalysis(pf_ts_trans, [300,400,500,600])
    ta_ts_trans.write_to_file("%s_thermo_ts_trans.csv" % prefix)

    # Define kinetic models for the chemical reaction. These are the mandatory arguments:
    #  1) a list of reactant partition functions
    #     (one for unimolecular, two for bimolecular, ...)
    #  2) the transition state partition function
    # There are two more optional arguments
    #  3) cp: model at constant pressure, default=True
    #  4) tunneling: a model for the tunelling correction
    km_gauche = KineticModel([pf_ethyl, pf_ethene], pf_ts_gauche)
    km_trans = KineticModel([pf_ethyl, pf_ethene], pf_ts_trans)

    # Analyze the chemical reactions. These are the arguments:
    #  1) A kinetic model
    #  2) the starting temperature for the fit
    #  3) the final temperature for the fit
    # The following argument is optional:
    #  4) temp_step: The interval on the temperature grid in Kelvin, 10 is default
    ra_gauche = ReactionAnalysis(km_gauche, 300, 600)
    ra_trans = ReactionAnalysis(km_trans, 300, 600)


    # make the Arrhenius plots
    pt.clf()
    ra_gauche.plot_arrhenius(label="gauche", color="red")
    ra_trans.plot_arrhenius(label="trans", color="blue")
    pt.legend(loc=0)
    pt.savefig("%s_arrhenius.png" % prefix)

    # Estimate the error on the kinetic parameters due to level of theory artifacts
    # with Monte Carlo sampling. The monte_carlo method takes three optional
    # arguments:
    #  1) freq_error: the absolute stochastic error on the frequencies (default=1*invcm)
    #  2) energy_error: the absolute error on the energy (default=0.0)
    #  3) num_iter: the number of monte carlo samples (default=100)
    ra_gauche.monte_carlo(num_iter=100)
    ra_trans.monte_carlo(num_iter=100)
    # plot the parameters, this includes the monte carlo results
    pt.clf()
    ra_gauche.plot_parameters(label="gauche", color="red")
    ra_trans.plot_parameters(label="trans", color="blue")
    pt.legend(loc=0)
    pt.savefig("%s_parameters.png" % prefix)
    # write all results to a file.
    ra_gauche.write_to_file("%s_reaction_gauche.txt" % prefix)
    ra_trans.write_to_file("%s_reaction_trans.txt" % prefix)

    def write_ra_summary(fn, ra):
        f = file(fn, "w")
        print >> f, "% 10.5e % 10.5e % 10.5e % 10.5e    %10.5e %10.2e    %10.2e %10.2e" % (
            ra.compute_rate_coeff(300)/ra.unit,
            ra.compute_rate_coeff(400)/ra.unit,
            ra.compute_rate_coeff(500)/ra.unit,
            ra.compute_rate_coeff(600)/ra.unit,
            ra.A/ra.unit,
            ra.Ea/kjmol,
            ra.compute_delta_G(0.0)/kjmol,
            ra.compute_delta_E()/kjmol,
        )
        f.close()

    write_ra_summary("%s_summary_gauche.txt" % prefix, ra_gauche)
    write_ra_summary("%s_summary_trans.txt" % prefix, ra_trans)


usage = """USAGE: ./reaction.py dirname

Analyzes the kinetics of the addition of ethene to ethyl with and without
hindered rotors. It looks at both the trans and gauce pathways.
"""


def main():
    from optparse import OptionParser
    parser = OptionParser(usage)
    options, args = parser.parse_args()

    if len(args) != 1:
        parser.error("Expecting exactly on argument")

    print "Processing %s" % args[0]
    os.chdir(args[0])
    load_sp = args[0].startswith("GEO")
    for do_rotor in True, False:
        for do_counterpoise in True, False:
            #try:
                run(do_rotor, do_counterpoise, load_sp)
                print "      OK: do_rotor=%i, do_counterpoise=%i, load_sp=%i" % (do_rotor, do_counterpoise, load_sp)
            #except (IOError, KeyError), e:
            #    print "  Failed: do_rotor=%i, do_counterpoise=%i, load_sp=%i" % (do_rotor, do_counterpoise, load_sp)
            #    print e


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin library.
from tamkin import *
# import matplotlib.pyplot as pt for plotting
import matplotlib.pyplot as pt


# Define an auxiliary function
def load_rotor(mol, filename, rotsym, even, top_indexes=None):
    rot_scan = load_rotscan_g03log(filename, top_indexes)
    rotor = Rotor(rot_scan, mol, rotsym=rotsym, even=even)
    return rotor

mol_ethyl = load_molecule_g03fchk("ethyl/freq/gaussian.fchk")
mol_ethene = load_molecule_g03fchk("ethene/freq/gaussian.fchk")
mol_ts_gauche = load_molecule_g03fchk("ts_ad1/freq_gauche/gaussian.fchk")
mol_ts_trans = load_molecule_g03fchk("ts_ad1/freq_trans/gaussian.fchk")
# Perform normal mode analysis on the molecules
nma_ethyl = NMA(mol_ethyl, ConstrainExt())
nma_ethene = NMA(mol_ethene, ConstrainExt())
nma_ts_gauche = NMA(mol_ts_gauche, ConstrainExt())
nma_ts_trans = NMA(mol_ts_trans, ConstrainExt())
# Construct the rotors
rotor_ethyl = load_rotor(mol_ethyl, "ethyl/scan/gaussian.log", 6, True)
rotor1_ts_gauche = load_rotor(mol_ts_gauche, "ts_ad1/scan1/gaussian.log", 3, False, [4,5,6])
rotor2_ts_gauche = load_rotor(mol_ts_gauche, "ts_ad1/scan2/gaussian.log", 1, True)
rotor1_ts_trans = load_rotor(mol_ts_trans, "ts_ad1/scan1/gaussian.log", 3, False, [4,5,6])
rotor2_ts_trans = load_rotor(mol_ts_trans, "ts_ad1/scan2/gaussian.log", 1, True)
# Construct the partition functions.
pf_ethyl = PartFun(nma_ethyl, [
    ExtTrans(), ExtRot(), Electronic(2),
    Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
    rotor_ethyl,
])
pf_ethene = PartFun(nma_ethene, [
    ExtTrans(), ExtRot(),
    Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
])
pf_ts_gauche = PartFun(nma_ts_gauche, [
    ExtTrans(), ExtRot(), Electronic(2),
    Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
    rotor1_ts_gauche, rotor2_ts_gauche,
])
pf_ts_trans = PartFun(nma_ts_trans, [
    ExtTrans(), ExtRot(), Electronic(2),
    Vibrations(freq_scaling=0.9614, zp_scaling=0.9806),
    rotor1_ts_trans, rotor2_ts_trans,
])

# Analyse the partition functions in detail
ta_ethyl = ThermoAnalysis(pf_ethyl, [300,400,500,600])
ta_ethyl.write_to_file("thermo_ethyl.csv")
ta_ethene = ThermoAnalysis(pf_ethene, [300,400,500,600])
ta_ethene.write_to_file("thermo_ethene.csv")
ta_ts_gauche = ThermoAnalysis(pf_ts_gauche, [300,400,500,600])
ta_ts_gauche.write_to_file("thermo_ts_gauche.csv")
ta_ts_trans = ThermoAnalysis(pf_ts_trans, [300,400,500,600])
ta_ts_trans.write_to_file("thermo_ts_trans.csv")

# Define kinetic models for the chemical reaction. These are the mandatory arguments:
#  1) a list of reactant partition functions
#     (one for unimolecular, two for bimolecular, ...)
#  2) the transition state partition function
# There is one more optional argument. (not used here).
#  3) tunneling: a model for the tunelling correction
km_gauche = KineticModel([pf_ethyl, pf_ethene], pf_ts_gauche)
km_trans = KineticModel([pf_ethyl, pf_ethene], pf_ts_trans)

# Analyze the chemical reactions. These are the arguments:
#  1) A kinetic model
#  2) the starting temperature for the fit
#  3) the final temperature for the fit
# The following argument is optional:
#  4) temp_step: The interval on the temperature grid in Kelvin, 10 is default
ra_gauche = ReactionAnalysis(km_gauche, 300, 600)
ra_trans = ReactionAnalysis(km_trans, 300, 600)

# make the Arrhenius plots
pt.clf()
ra_gauche.plot_arrhenius(label="gauche", color="red")
ra_trans.plot_arrhenius(label="trans", color="blue")
pt.legend(loc=0)
pt.savefig("arrhenius.png")

# Estimate the error on the kinetic parameters due to level of theory artifacts
# with Monte Carlo sampling. The monte_carlo method takes three optional
# arguments:
#  1) freq_error: the absolute stochastic error on the frequencies (default=1*invcm)
#  2) energy_error: the absolute error on the energy (default=0.0)
#  3) num_iter: the number of monte carlo samples (default=100)
ra_gauche.monte_carlo(num_iter=10)
ra_trans.monte_carlo(num_iter=10)
# plot the parameters, this includes the monte carlo results
pt.clf()
ra_gauche.plot_parameters(label="gauche", color="red")
ra_trans.plot_parameters(label="trans", color="blue", marker="^")
pt.legend(loc=0, numpoints=1)
pt.xlabel("E$_a$ [kJ mol$^{-1}$]")
pt.ylabel("ln(A) [ln(m$^3$ mol$^{-1}$ s$^{-1}$)]")
pt.xlim(33.24,33.75)
pt.savefig("parameters.png")
# write all results to a file.
ra_gauche.write_to_file("reaction_gauche.txt")
ra_trans.write_to_file("reaction_trans.txt")
# Plot the energy levels and the potential of the hindered rotor. The
# temperature argument is used to indicate the population of each level in the
# plot.
rotor_ethyl.plot_levels("rotor_ethyl_energy_levels.png", 300)
rotor1_ts_gauche.plot_levels("rotor1_ts_gauche_energy_levels.png", 300)
rotor2_ts_gauche.plot_levels("rotor2_ts_gauche_energy_levels.png", 300)
rotor1_ts_trans.plot_levels("rotor1_ts_trans_energy_levels.png", 300)
rotor2_ts_trans.plot_levels("rotor2_ts_trans_energy_levels.png", 300)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin library.
from tamkin import *

# load the fixed atoms
#fixed_atoms = load_fixed_g03com("Zp_p_react.14mei.com")
fixed_react = load_fixed_g03com("Zp_p_react.28aug.com")
fixed_trans = load_fixed_g03com("Zp_p_TS.28aug.com")
# load the gaussian data
mol_react = load_molecule_g03fchk("Zp_p_react.28aug.fchk", "Zp_p_react.14mei.fchk")
mol_trans = load_molecule_g03fchk("Zp_p_TS.28aug.fchk", "5Tp_p_TS.oniom21apr_HF.fchk")
# Construct the three partition functions.
pf_react = PartFun(NMA(mol_react, PHVA(fixed_react)))
pf_trans = PartFun(NMA(mol_trans, PHVA(fixed_trans)))


# Define a kinetic model for the chemical reaction. These are the mandatory arguments:
#  1) a list of reactant partition functions
#     (one for unimolecular, two for bimolecular, ...)
#  2) the transition state partition function
# There is one more optional argument:
#  3) tunneling: a model for the tunelling correction
km = KineticModel([pf_react], pf_trans, tunneling=None)

# Analyze the chemical reaction. These are the arguments:
#  1) A kinetic model
#  2) the starting temperature for the fit
#  3) the final temperature for the fit
# The following argument is optional:
#  4) temp_step: The interval on the temperature grid in Kelvin, 10 is default
ra = ReactionAnalysis(km, 670, 770)
ra.plot_arrhenius("arrhenius.png") # make the Arrhenius plot

# Estimate the error on the kinetic parameters due to level of theory artifacts
# with Monte Carlo sampling. The monte_carlo method takes three optional
# arguments:
#  1) freq_error: the absolute stochastic error on the frequencies (default=1*invcm)
#  2) energy_error: the absolute error on the energy (default=0.0)
#  3) num_iter: the number of monte carlo samples (default=100)
ra.monte_carlo()
# plot the parameters, this includes the monte carlo results
ra.plot_parameters("parameters.png")
# write all results to a file.
ra.write_to_file("reaction.txt") # summary of the analysis



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from tamkin.data import Molecule

from molmod.periodic import periodic
from molmod.unit_cells import UnitCell
from molmod.units import angstrom, amu

import numpy as np


__all__ = ["load_molecule_cp2k", 'load_fixed_cp2k']


def load_molecule_cp2k(fn_sp, fn_freq, multiplicity=1, is_periodic=True):
    """Load a molecule with the Hessian from a CP2K computation

       Arguments:
        | fn_sp   --  The filename of the single point .out file containing the
                      energy and the forces.
        | fn_freq  --  The filename of the frequency .out file containing the
                       hessian

       Optional arguments:
        | multiplicity  --  The spin multiplicity of the electronic system
                            [default=1]
        | is_periodic  --  True when the system is periodic in three dimensions.
                           False when the systen is aperiodic. [default=True]
        | unit_cell  --  The unit cell vectors for periodic structures
    """
    # auxiliary routine to read atoms
    def atom_helper(f):
        # skip some lines
        for i in xrange(3):
            f.readline()
        # read the atom lines until an empty line is encountered
        numbers = []
        coordinates = []
        masses = []
        while True:
            line = f.readline()
            if len(line.strip()) == 0:
                break
            symbol = line[14:19].strip()[:2]
            atom = periodic[symbol]
            if atom is None:
                symbol = symbol[:1]
                atom = periodic[symbol]
            if atom is None:
                numbers.append(0)
            else:
                numbers.append(atom.number)
            coordinates.append([float(line[22:33]), float(line[34:45]), float(line[46:57])])
            masses.append(float(line[72:]))

        numbers = np.array(numbers)
        coordinates = np.array(coordinates)*angstrom
        masses = np.array(masses)*amu
        return numbers, coordinates, masses


    # auxiliary routine to read forces
    def force_helper(f, skip, offset):
        # skip some lines
        for i in xrange(skip):
            f.readline()
        # Read the actual forces
        tmp = []
        while True:
            line = f.readline()
            if line == "\n":
                break
            if line == "":
                raise IOError("End of file while reading gradient (forces).")
            words = line.split()
            try:
                tmp.append([float(words[offset]), float(words[offset+1]), float(words[offset+2])])
            except StandardError:
                break
        return -np.array(tmp) # force to gradient

    # go through the single point file: energy and gradient
    energy = None
    gradient = None
    with open(fn_sp) as f:
        while True:
            line = f.readline()
            if line == "":
                break
            if line.startswith(" ENERGY|"):
                energy = float(line[58:])
            elif line.startswith(" MODULE") and "ATOMIC COORDINATES" in line:
                numbers, coordinates, masses = atom_helper(f)
            elif line.startswith(" FORCES|"):
                gradient = force_helper(f, 0, 1)
                break
            elif line.startswith(' ATOMIC FORCES in [a.u.]'):
                gradient = force_helper(f, 2, 3)
                break
        if energy is None or gradient is None:
            raise IOError("Could not read energy and/or gradient (forces) from single point file.")

    # go through the freq file: lattic vectors and hessian
    with open(fn_freq) as f:
        vectors = np.zeros((3, 3), float)
        for line in f:
            if line.startswith(" CELL"):
                break
        for axis in range(3):
            line = f.next()
            vectors[:,axis] = np.array( [float(line[29:39]), float(line[39:49]), float(line[49:59])] )
        unit_cell = UnitCell(vectors*angstrom)

        free_indices = _load_free_low(f)
        if len(free_indices) > 0:
            total_size = coordinates.size
            free_size = len(free_indices)
            hessian = np.zeros((total_size, total_size), float)
            i2 = 0
            while i2 < free_size:
                num_cols = min(5, free_size - i2)
                f.next() # skip two lines
                f.next()
                for j in xrange(free_size):
                    line = f.next()
                    words = line.split()
                    for i1 in xrange(num_cols):
                        hessian[free_indices[i2 + i1], free_indices[j]] = \
                            float(words[i1 + 2])
                i2 += num_cols
        else:
            raise IOError("Could not read hessian from freq file.")

    # symmetrize
    hessian = 0.5*(hessian+hessian.transpose())
    # cp2k prints a transformed hessian, here we convert it back to the normal
    # hessian in atomic units.
    conv = 1e-3*np.array([masses, masses, masses]).transpose().ravel()**0.5
    hessian *= conv
    hessian *= conv.reshape((-1,1))

    return Molecule(
        numbers, coordinates, masses, energy, gradient,
        hessian, multiplicity, 0, is_periodic, unit_cell=unit_cell
    )


def _load_free_low(f):
    '''Helper function to load the indices of the free atoms from a CP2K freq file

    Parameters
    ----------
    f: open file
       An open file object with the CP2K freq output.

    Returns
    -------
    free_indices: list of int
       A list of integer indexes of the free rows and columbs of the Hessian matrix.
    '''
    free_indices = []
    for line in f:
        if line.startswith(" VIB| Vibrational Analysis Info"):
            break
    for line in f:
        if line.startswith(" VIB| REPLICA Nr."):
            words = line.split()
            if words[-2] == '+':
                free_index = 3*(int(words[-5])-1)
                if words[-3] == 'Y':
                    free_index += 1
                elif words[-3] == 'Z':
                    free_index += 2
                free_indices.append(free_index)
        if line.startswith(" VIB| Hessian in cartesian coordinates"):
            break
    return free_indices


def load_fixed_cp2k(fn_freq):
    '''Load the fixed atoms from a CP2K freq output

    Parameters
    ----------
    fn_freq: str
             The filename of the CP2K freq output

    Returns
    -------
    fixed: numpy bool array
           The length of the vector corresponds to the number of atoms. For every fixed
           atom, the vector element is True. False otherwise.
    '''
    with open(fn_freq) as f:
        natom = None
        for line in f:
            if line.startswith('                             - Atoms: '):
                natom = int(line.split()[-1])
                break
        free_indices = _load_free_low(f)
    if natom is None:
        raise IOError('Could not read number of atoms from CP2K output.')
    if len(free_indices) == 0:
        raise IOError('Could not find the free atoms.')
    free_atoms = np.array(free_indices[::3])/3
    return np.array([i for i in xrange(natom) if i not in free_atoms])



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Partition functions based on the harmonic oscillator approximation.

   The workhorse of this module is the **PartFun** class. A PartFun object
   represents a partition function with a solid interface. PartFun objects
   can be used to study chemical equilibrium, rate coefficients and various
   thermodynamic properties.

   These are the other classes and functions present in this module:

   * **Abstract classes:**
       * Info
       * StatFys
       * StatFysTerms
   * **Contributions to the partition function:**
       * Electronic
       * ExtTrans
       * ExtRot
       * Vibrations
       * Rotor (see rotor.py)
   * **Helper functions:**
       * helper_levels, helpert_levels, helpertt_levels,
       * helper_vibrations, helpert_vibrations, helpertt_vibrations

   **Important**: Partition functions can be constructed for NpT gases, NVT
   gases and many other systems. The return values of methods such as
   ``free_energy``, ``internal_heat`` and ``heat_capacity`` are often given
   specialized names in the context of different partition functions. For
   example, chemists tend to use the following names:

   * 3D NVT gas:
       - ``PartFun.free_energy`` -> the Helmholtz free energy
       - ``PartFun.internal_heat`` -> the internal energy
       - ``PartFun.heat_capacity`` -> the heat capacity at constant volume

   * 3D NpT gas:
       - ``PartFun.free_energy`` -> the Gibbs free energy
       - ``PartFun.internal_heat`` -> the enthalpy
       - ``PartFun.heat_capacity`` -> the heat capacity at constant pressure

   Don't say we did not warn you. Terminology can be very confusing.

   All the extensive thermodynamic quantities computed here are in atomic units
   per molecule. If you want to express the Gibbs free energy at 300 Kelvin of a
   system in kJ/mol, use the molmod module to perform unit conversions. For
   example::

     >>> pf = PartFun(..., [ExtTrans(cp=True)])
     >>> print pf.free_energy(300)/kjmol
"""


from molmod import boltzmann, lightspeed, atm, bar, amu, centimeter, kjmol, \
    planck, mol, meter, newton

import numpy as np


__all__ = [
    "Info", "StatFys", "StatFysTerms",
    "helper_levels", "helpert_levels", "helpertt_levels",
    "Electronic", "ExtTrans", "ExtRot", "PCMCorrection",
    "Vibrations",
    "helper_vibrations", "helpert_vibrations", "helpertt_vibrations",
    "PartFun",
]


class Info(object):
    """An object that has a name and that can dump info to a file."""
    def __init__(self, name):
        """
           Arguments:
            | ``name`` -- the name used for this object in the output
        """
        self.name = name

    def dump(self, f):
        """Write a description to file.

           Arguments:
            | ``f`` -- the file object to write to
        """
        print >> f, "  %s" % self.name.upper()

    def dump_values(self, f, label, values, format, num_col=8):
        """Write a nicely formatted array of numbers to file.

           Arguments:
            | ``f`` -- the file object to write to
            | ``label`` -- a label that explains the meaning of the numbers
            | ``values`` -- the array with numbers
            | ``format`` -- a Python format string for one number

           Optional argumet:
            | ``num_col`` -- the number of columns [default=8]
        """
        parts = ["    "]
        for counter, value in enumerate(values):
            parts.append(format % value)
            if counter % num_col == num_col-1:
                parts.append("\n    ")
        if len(parts) > 1:
            if parts[-1] == "\n    ":
                parts.pop()
            parts.insert(0, "    %s:\n" % label)
            print >> f, "".join(parts)


class StatFys(object):
    """Abstract class for (contributions to) the parition function.

       The constructor (__init__) and four methods (init_part_fun, helper,
       helpert, helpertt) must be implemented in derived classes.
    """
    def init_part_fun(self, nma, partf):
        """Compute parameters that depend on nma and partition function.

           Arguments:
            | ``nma`` -- an NMA object
            | ``partf`` -- A PartFun object

           This method is called by the PartFun object and should not be called
           by the user. When the PartFun object is initialized, each
           contribution can further initialize its parameters based on the
           information available in the nma and partf objects.
        """
        pass

    def helper(self, temp, n):
        r"""Helper function.

           Returns

           .. math:: T^n \frac{\ln(Z_N)}{N}

           where :math:`Z_N` is (the contribution to) the many body partition
           function and :math:`N` is the total number of particles. In all
           cases, except for the translational contribution, this comes down to:

           .. math:: T^n \ln(Z_1)

           where :math:`\ln(Z_1)` is the single-particle contribution to the
           partition function.

           Arguments:
            | ``temp`` -- the temperature
            | ``n`` -- the power for the temperature factor
        """
        raise NotImplementedError

    def helpert(self, temp, n):
        r"""Helper function T.

           Returns

           .. math:: T^n \frac{d \left(\frac{\ln(Z_N)}{N}\right)}{dT}

           where :math:`Z_N` is (the contribution to) the many body partition
           function and :math:`N` is the total number of particles.  In all
           cases, except for the translational contribution, this comes down to:

           .. math:: T^n \frac{\partial \ln(Z_1)}{\partial T}

           where :math:`\ln(Z_1)` is the single-particle contribution to the
           partition function.

           Arguments:
            | ``temp`` -- the temperature
            | ``n`` -- the power for the temperature factor
        """
        raise NotImplementedError

    def helpertt(self, temp, n):
        r"""Helper function TT.

           Returns

           .. math:: T^n \frac{d^2 \left(\frac{\ln(Z_N)}{N}\right)}{dT^2}

           where :math:`Z_N` is (the contribution to) the many body partition
           function and :math:`N` is the total number of particles. In all
           cases, except for the translational contribution, this comes down to:

           .. math:: T^n \frac{\partial^2 \ln(Z_1)}{\partial T^2}

           where :math:`\ln(Z_1)` is the single-particle contribution to the
           partition function.

           Arguments:
            | ``temp`` -- the temperature
            | ``n`` -- the power for the temperature factor
        """
        raise NotImplementedError

    def helpern(self, temp, n):
        r"""Helper function N.

           Returns

           .. math:: T^n \frac{\partial \ln(Z_N)}{\partial N}

           where :math:`Z_N` is (the contribution to) the many body partition
           function and :math:`N` is the total number of particles. This is used
           to compute the chemical potential and reaction free energies. In all
           cases, except for the total partition function and translational
           contribution, this comes down to the method ``helper``.

           Arguments:
            | ``temp`` -- the temperature
            | ``n`` -- the power for the temperature factor
        """
        # By default return helper
        return self.helper(temp, n)

    def helperv(self, temp, n):
        r"""Helper function V

           This always the same as the method ``helpern``, except for the
           total partition function and the translational contribution to the
           partition function. Then it returns

           .. math:: T^n \frac{\partial \ln(Z_N)}{\partial N} - \ln\left(\frac{V}{N}\right)

           This is used for the computation of equilibrium constants, and rate
           constants.

           Arguments:
            | ``temp`` -- temperature
            | ``n`` -- the power for the temperature factor
        """
        # By default return helper
        return self.helper(temp, n)

    def log(self, temp, helper=None):
        r"""Log function

           The logarithm of the N-particle partition function divided by the
           number of particles:

           .. math:: \frac{\ln(Z_N)}{N}

           For all contributions, except for the translational, this comes down
           to:

           .. math:: \ln(Z_1)

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helper`` -- an alternative implementation of helper
                            [default=self.helper]
        """
        if helper is None:
            helper = self.helper
        return helper(temp, 0)

    def logt(self, temp, helpert=None):
        r"""Log function T

           The derivative towards temperature of the logarithm of the N-particle
           partition function divided by the number of particles:

           .. math:: \frac{d \left(\frac{\ln(Z_N)}{N}\right)}{dT}

           For all contributions, except for the translational, this comes down
           to:

           .. math:: \frac{\partial \ln(Z_1)}{\partial T}

           Argument:
            | ``temp`` -- the temperature

           Optional arguments:
            | ``helpert`` -- an alternative implementation of helpert
                             [default=self.helpert]
        """
        if helpert is None:
            helpert = self.helpert
        return helpert(temp, 0)

    def logtt(self, temp, helpertt=None):
        r"""Log function TT

           The second derivative towards temperature of the logarithm of the
           N-particle partition function divided by the number of particles.

           .. math:: \frac{d^2 \left(\frac{\ln(Z_N)}{N}\right)}{dT^2}

           For all contributions, except for the translational, this comes down
           to:

           .. math:: \frac{\partial^2 \ln(Z_1)}{\partial T^2}

           Argument:
            | ``temp`` -- the temperature

           Optional arguments:
            | ``helpertt`` -- an alternative implementation of helpertt
                              [default=self.helpertt]
        """
        if helpertt is None:
            helpertt = self.helpertt
        return helpertt(temp, 0)

    def logn(self, temp, helpern=None):
        r"""Log function N

           The derivative of the logarithm of the many-particle partition
           function towards to number of particles.

           .. math:: \frac{\partial \ln(Z_N)}{\partial N}

           In most cases, this is the same as the logarithm of the
           single-particle partition divided by the number of particles. (see
           method ``log``.) The only exception is the translational contribution
           to the partition function, and hence also the total partition
           function.

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helpern`` -- an alternative implementation of helpern
                             [default=self.helpern]
        """
        if helpern is None:
            helpern = self.helpern
        return helpern(temp, 0)

    def logv(self, temp, helperv=None):
        r"""Log function V

           This always the same as the method ``logn``, except for the
           total partition function and the translational contribution to the
           partition function. Then it returns

           .. math:: \frac{\partial \ln(Z_N)}{\partial N} - \ln\left(\frac{V}{N}\right)

           This is used for the computation of equilibrium constants, and rate
           constants.

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helperv`` -- an alternative implementation of helperv
                             [default=self.helperv]
        """
        if helperv is None:
            helperv = self.helperv
        return helperv(temp, 0)

    def internal_heat(self, temp, helpert=None):
        """Computes the internal heat per molecule.

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helpert`` -- an alternative implementation of helpert
                             [default=self.helpert]
        """
        if helpert is None:
            helpert = self.helpert
        return boltzmann*helpert(temp, 2)

    def heat_capacity(self, temp, helpert=None, helpertt=None):
        """Computes the heat capacity per molecule.

           Argument:
            | ``temp`` -- the temperature

           Optional arguments:
            | ``helpert`` -- an alternative implementation of helpert
                             [default=self.helpert]
            | ``helpertt`` -- an alternative implementation of helpertt
                              [default=self.helpertt]
        """
        if helpert is None:
            helpert = self.helpert
        if helpertt is None:
            helpertt = self.helpertt
        return boltzmann*(2*helpert(temp, 1) + helpertt(temp, 2))

    def entropy(self, temp, helper=None, helpert=None):
        """Computes the entropy contribution per molecule.

           Argument:
            | ``temp`` -- the temperature

           Optional arguments:
            | ``helper`` -- an alternative implementation of helper
                            [default=self.helper]
            | ``helpert`` -- an alternative implementation of helpert
                             [default=self.helpert]
        """
        if helper is None:
            helper = self.helper
        if helpert is None:
            helpert = self.helpert
        return boltzmann*(helper(temp, 0) + helpert(temp, 1))

    def free_energy(self, temp, helper=None):
        """Computes the free energy per molecule.

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helper`` -- an alternative implementation of helper
                            [default=self.helper]
        """
        if helper is None:
            helper = self.helper
        return -boltzmann*helper(temp, 1)

    def zero_point_energy(self, helpern=None):
        """Return the zero-point energy.

           Optional argument:
            | ``helper`` -- an alternative implementation of helpern
                            [default=self.helpern]

           In TAMkin the zero-point energy is defined as the limit of the
           chemical potential for the temperature going towards zero.
        """
        return self.chemical_potential(0, helpern)

    def chemical_potential(self, temp, helpern=None):
        """Computes the chemical potential.

           Argument:
            | ``temp`` -- the temperature

           Optional argument:
            | ``helper`` -- an alternative implementation of helpern
                            [default=self.helpern]
        """
        if helpern is None:
            helpern = self.helpern
        return -boltzmann*helpern(temp, 1)


class StatFysTerms(StatFys):
    """Abstract class for (contributions to) the parition function with multiple terms.

       The different terms (or factors if you like) are of the same mathematical
       structure.

       The constructor (__init__) and the four methods (init_part_fun,
       helper_terms, helpert_terms, helpertt_terms) must be implemented in
       derived classes.
    """
    def __init__(self, num_terms):
        """
           Arguments:
             num_terms`` -- the number of terms present in this contribution.
        """
        self.num_terms = num_terms

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        return self.helper_terms(temp, n).sum()

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        return self.helpert_terms(temp, n).sum()

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        return self.helpertt_terms(temp, n).sum()

    def helpern(self, temp, n):
        """See :meth:`StatFys.helpern`."""
        return self.helpern_terms(temp, n).sum()

    def helperv(self, temp, n):
        """See :meth:`StatFys.helperv`."""
        return self.helperv_terms(temp, n).sum()

    def helper_terms(self, temp, n):
        """Returns an array with all the helper results for the distinct terms.

           This is just an array version of :meth:`StatFys.helper`.
        """
        raise NotImplementedError

    def helpert_terms(self, temp, n):
        """Returns an array with all the helpert results for the distinct terms.

           This is just an array version of :meth:`StatFys.helpert`.
        """
        raise NotImplementedError

    def helpertt_terms(self, temp, n):
        """Returns an array with all the helpertt results for the distinct terms.

           This is just an array version of :meth:`StatFys.helpertt`.
        """
        raise NotImplementedError

    def helpern_terms(self, temp, n):
        """Returns an array with all the helpern results for the distinct terms.

           This is just an array version of :meth:`StatFys.helpern`.
        """
        # by default, this is the same as helper_terms
        return self.helper_terms(temp, n)

    def helperv_terms(self, temp, n):
        """Returns an array with all the helperv results for the distinct terms.

           This is just an array version of :meth:`StatFys.helperv`.
        """
        # by default, this is the same as helper_terms
        return self.helper_terms(temp, n)

    def log_terms(self, temp):
        """Returns an array with log results for the distinct terms.

           This is just an array version of :meth:`StatFys.log`.
        """
        return self.log(temp, self.helper_terms)

    def logt_terms(self, temp):
        """Returns an array with logt results for the distinct terms.

           This is just an array version of :meth:`StatFys.logt`.
        """
        return self.logt(temp, self.helpert_terms)

    def logtt_terms(self, temp):
        """Returns an array with logtt results for the distinct terms.

           This is just an array version of :meth:`StatFys.logtt`.
        """
        return self.logtt(temp, self.helpertt_terms)

    def logv_terms(self, temp):
        """Returns an array with logv results for the distinct terms.

           This is just an array version of :meth:`StatFys.logv`.
        """
        return self.logv(temp, self.helperv_terms)

    def logn_terms(self, temp):
        """Returns an array with logn results for the distinct terms.

           This is just an array version of :meth:`StatFys.logn`.
        """
        return self.logn(temp, self.helperv_terms)

    def internal_heat_terms(self, temp):
        """Returns an array with internal_heat results for the distinct terms.

           This is just an array version of :meth:`StatFys.internal_heat`.
        """
        return self.internal_heat(temp, self.helpert_terms)

    def heat_capacity_terms(self, temp):
        """Returns an array with heat_capacity results for the distinct terms.

           This is just an array version of :meth:`StatFys.heat_capacity`.
        """
        return self.heat_capacity(temp, self.helpert_terms, self.helpertt_terms)

    def entropy_terms(self, temp):
        """Returns an array with entropy results for the distinct terms.

           This is just an array version of :meth:`StatFys.entropy`.
        """
        return self.entropy(temp, self.helper_terms, self.helpert_terms)

    def free_energy_terms(self, temp):
        """Returns an array with free_energy results for the distinct terms.

           This is just an array version of :meth:`StatFys.free_energy`.
        """
        return self.free_energy(temp, self.helper_terms)

    def zero_point_energy_terms(self):
        """Returns an array with zero_point_energy results for the distinct terms.

           This is just an array version of :meth:`StatFys.chemical_potential`.
        """
        return self.zero_point_energy(self.helpern_terms)

    def chemical_potential_terms(self, temp):
        """Returns an array with chemical_potential results for the distinct terms.

           This is just an array version of :meth:`StatFys.chemical_potential`.
        """
        return self.chemical_potential(temp, self.helpern_terms)


def _check_levels(temp, bfs, Z):
    if bfs[bfs.argmin()]/Z > 0.01:
        raise ValueError('The highest energy level is occupied by more than 1%.')


def helper_levels(temp, n, energy_levels, check=False):
    """Helper 0 function for a system with the given energy levels.

       Returns T^n ln(Z), where Z is the partition function

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``energy_levels`` -- an array with energy levels

       Optional argument:
        | ``check`` -- when set to True, an error is raise when the highest
                       energy level is occupied by more than 1% the the current
                       temperature.
    """
    # this is defined as a function because multiple classes need it
    if temp == 0:
        energy = energy_levels[0]
        degeneracy = 1
        while (degeneracy<len(energy_levels) and energy_levels[0]==energy_levels[degeneracy]):
            degeneracy += 1
        return float(temp)**n*np.log(degeneracy) - float(temp)**(n-1)*energy
    else:
        bfs = np.exp(-energy_levels/(boltzmann*temp))
        Z = bfs.sum()
        if check:
            _check_levels(temp, bfs, Z)
        return float(temp)**n*np.log(Z)

def helpert_levels(temp, n, energy_levels, check=False):
    """Helper 1 function for a system with the given energy levels.

       Returns T^n (d ln(Z) / dT), where Z is the partition function

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``energy_levels`` -- an array with energy levels

       Optional argument:
        | ``check`` -- when set to True, an error is raise when the highest
                       energy level is occupied by more than 1% the the current
                       temperature.
    """
    # this is defined as a function because multiple classes need it
    if temp == 0:
        raise NotImplementedError
    else:
        bfs = np.exp(-energy_levels/(boltzmann*temp))
        Z = bfs.sum()
        if check:
            _check_levels(temp, bfs, Z)
        return float(temp)**(n-2)*(bfs*energy_levels).sum()/Z/boltzmann

def helpertt_levels(temp, n, energy_levels, check=False):
    """Helper 2 function for a system with the given energy levels.

       Returns T^n (d^2 ln(Z) / dT^2), where Z is the partition function

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``energy_levels`` -- an array with energy levels

       Optional argument:
        | ``check`` -- when set to True, an error is raise when the highest
                       energy level is occupied by more than 1% the the current
                       temperature.
    """
    # this is defined as a function because multiple classes need it
    if temp == 0:
        raise NotImplementedError
    else:
        bfs = np.exp(-energy_levels/(boltzmann*temp))
        Z = bfs.sum()
        if check:
            _check_levels(temp, bfs, Z)
        return float(temp)**(n-4)/boltzmann**2*((bfs*energy_levels**2).sum()/Z) \
               -2*float(temp)**(n-3)/boltzmann*((bfs*energy_levels).sum()/Z) \
               -float(temp)**(n-4)/boltzmann**2*((bfs*energy_levels).sum()/Z)**2


class Electronic(Info, StatFys):
    """The electronic contribution to the partition function."""
    # TODO: this should also include the potential energy from the ab initio
    # computation. This is now added in the PartFun object.
    def __init__(self, multiplicity=None):
        """
           Optional argument:
            | ``multiplicity`` -- the spin multiplicity of the electronic system

           When the optional argument is not given, it is determined when the
           PartFun object is constructed.
        """
        self.multiplicity = multiplicity
        self.energy = None
        Info.__init__(self, "electronic")

    def init_part_fun(self, nma, partf):
        """See :meth:`StatFys.init_part_fun`."""
        if self.multiplicity is None:
            self.multiplicity = nma.multiplicity
            if self.multiplicity is None:
                raise ValueError("Spin multiplicity is not defined.")
        self.energy = nma.energy

    def dump(self, f):
        """See :meth:`Info.dump`."""
        Info.dump(self, f)
        print >> f, "    Multiplicity: %i" % self.multiplicity
        print >> f, "    Electronic energy: %.7f" % self.energy

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        result = float(temp)**n*np.log(self.multiplicity)
        if temp == 0.0:
            if n < 1:
                raise NotImplementedError
            else:
                result -= self.energy/boltzmann
        else:
            result -= float(temp)**(n-1)*self.energy/boltzmann
        return result

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        if temp == 0.0:
            if n < 2:
                raise NotImplementedError
            else:
                return self.energy/boltzmann
        else:
            return float(temp)**(n-2)*self.energy/boltzmann

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        if temp == 0.0:
            if n < 3:
                raise NotImplementedError
            else:
                return -2.0*self.energy/boltzmann
        else:
            return -2.0*float(temp)**(n-3)*self.energy/boltzmann


class ExtTrans(Info, StatFys):
    """The contribution from the external translation.

       This contribution includes many body terms and optional constant pressure
       corrections. It is based on the classical ideal gas approximation.
    """

    def __init__(self, cp=True, pressure=None, density=None, dim=3, mobile=None):
        """
           Optional arguments:
            | ``cp`` -- When True, an additional factor is included in the
                        partition function to model a constant pressure (or
                        constant surface tension) ensemble instead of a constant
                        volume (or constant surface) ensemble.
            | ``pressure`` -- (only allowed when cp==True)
                              The external pressure exerted on the system in
                              case of the NpT ensemble. The default is 1 atm
                              for 3D gases. The default for 2D systems is 75.64
                              miliNewton per meter, i.e. the surface tension of
                              water. For other dimensions, the default is 1.0.
            | ``density`` -- (only allowed when cp==False)
                             The density of the system in case of the NVT
                             ensemble. The default is 1.0 mol/meter**dim.
            | ``dim`` -- The dimension of the ideal gas.
            | ``mobile`` -- A list of atom indexes that are free to translate. In
                            case of a mobile molecule adsorbed on a surface, only
                            include atom indexes of the adsorbate. The default is
                            that all atoms are mobile.
        """
        if dim <= 0:
            raise ValueError("The dimension of the gas must be strictly positive.")
        if dim == 3:
            self.pressure_unit = bar
            self.pressure_unit_name = "bar"
        elif dim == 2:
            self.pressure_unit = 1e-3*newton/meter
            self.pressure_unit_name = "mN/m"
        else:
            self.pressure_unit = 1.0
            self.pressure_unit_name = "a.u."
        self.density_unit = mol/meter**dim
        self.density_unit_name = "mol*meter**%i" % (-dim)

        if cp:
            if density is not None:
                raise ValueError("The density can not be fixed in the NpT ensemble, i.e. it depends on the temperature.")
            if pressure is None:
                if dim == 3:
                    pressure = 1*atm
                elif dim == 2:
                    # approximately the surface tension of water:
                    pressure = self.pressure_unit*75.64
                else:
                    # whatever...
                    pressure = 1.0
            self._pressure = pressure
        else:
            if pressure is not None:
                raise ValueError("The pressure can not be fixed in the NVT ensemble, i.e. it depends on the temperature.")
            if density is None:
                density = self.density_unit
            self._density = density

        self._cp = cp
        self.dim = dim
        self._mobile = mobile
        Info.__init__(self, "translational")

    cp = property(lambda self: self._cp)
    mobile = property(lambda self: self._mobile)

    def _get_pressure(self):
        """The pressure in case of an NpT ensemble."""
        if self.cp:
            return self._pressure
        else:
            raise ValueError("The pressure is not a known constant in the NVT ensemble, i.e. it depends on the temperature.")

    def _set_pressure(self, pressure):
        """Set the pressure in case of an NpT ensemble.

           This will raise a ValueError in case this is a constant volume ensemble.
        """
        if self.cp:
            self._pressure = pressure
        else:
            raise ValueError("The pressure can not be fixed in the NVT ensemble, i.e. it depends on the temperature.")

    pressure = property(_get_pressure, _set_pressure)

    def _get_density(self):
        """The density in case of an NVT ensemble."""
        if not self.cp:
            return self._density
        else:
            raise ValueError("The density is not a known constant in the NpT ensemble, i.e. it depends on the temperature.")

    def _set_density(self, density):
        """Set the density in case of an NVT ensemble.

           This will raise a ValueError in case this is a constant pressure ensemble.
        """
        if not self.cp:
            self._density = density
        else:
            raise ValueError("The density can not be fixed in the NpT ensemble, i.e. it depends on the temperature.")

    density = property(_get_density, _set_density)

    def init_part_fun(self, nma, partf):
        """See :meth:`StatFys.init_part_fun`."""
        if self.mobile is None:
            self.mass = nma.mass
        else:
            self.mass = nma.masses[self.mobile].sum()

    def dump(self, f):
        """See :meth:`Info.dump`."""
        Info.dump(self, f)
        print >> f, "    Dimension: %i" % self.dim
        print >> f, "    Constant pressure: %s" % self.cp
        if self.cp:
            print >> f, "    Pressure [%s]: %.5f" % (self.pressure_unit_name, self._pressure/self.pressure_unit)
        else:
            print >> f, "    Density [%s]: %.5f" % (self.density_unit_name, self._density/self.density_unit)
        if self.cp:
            print >> f, "      BIG FAT WARNING!!!"
            print >> f, "      This is an NpT partition function."
            print >> f, "      Internal heat contains a PV term (and is therefore the enthalpy)."
            print >> f, "      Free energy contains a PV term (and is therefore the Gibbs free energy)."
            print >> f, "      The heat capacity is computed at constant pressure."
        else:
            print >> f, "      BIG FAT WARNING!!!"
            print >> f, "      This is an NVT partition function."
            print >> f, "      Internal heat does NOT contain a PV term."
            print >> f, "      Free energy does NOT contain a PV term (and is therefore the Helmholtz free energy)."
            print >> f, "      The heat capacity is computed at constant volume."
        print >> f, "    Mass [amu]: %f" % (self.mass/amu)

    def _z1(self, temp):
        return 0.5*self.dim*np.log(2*np.pi*self.mass*boltzmann*temp/planck**2)

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        if temp == 0:
            if n > 0:
                return 0.0
            else:
                raise NotImplementedError
        else:
            result = self._z1(temp)
            if self.cp:
                result += np.log(boltzmann*temp/self._pressure)
            else:
                result += 1.0 - np.log(self.density)
            return result*float(temp)**n

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        if temp == 0:
            raise NotImplementedError
        else:
            result = 0.5*self.dim
            if self.cp:
                result += 1
            return result*float(temp)**(n-1)

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        if temp == 0:
            raise NotImplementedError
        else:
            result = -0.5*self.dim
            if self.cp:
                result -= 1
            return result*float(temp)**(n-2)

    def helpern(self, temp, n):
        """See :meth:`StatFys.helpern`."""
        if temp == 0:
            if n > 0:
                return 0.0
            else:
                raise NotImplementedError
        else:
            result = self._z1(temp)
            if self.cp:
                result += np.log(boltzmann*temp/self._pressure)
            else:
                result += -np.log(self._density)
            return result*float(temp)**n

    def helperv(self, temp, n):
        r"""See :meth:`StatFys.helperv`."""
        if temp == 0:
            if n > 0:
                return 0.0
            else:
                raise NotImplementedError
        else:
            return self._z1(temp)*float(temp)**n


class ExtRot(Info, StatFys):
    """The contribution from the external rotation.

       This class is based on the integral approximation of the partition
       function.
    """
    def __init__(self, symmetry_number=None, im_threshold=1.0):
        """
           Optional arguments:
            | ``symmetry_number`` -- The rotational symmetry number of the
                                     molecule.
            | ``im_threshold``  --  When a moment of inertia drops below this
                                    threshold, it is discarded, which matters
                                    for linear molecules.
        """
        self.symmetry_number = symmetry_number
        self.im_threshold = im_threshold
        Info.__init__(self, "rotational")

    def init_part_fun(self, nma, partf):
        """See :meth:`StatFys.init_part_fun`."""
        if nma.periodic:
            raise ValueError("There is no external rotation in periodic systems.")
        self.inertia_tensor = nma.inertia_tensor
        self.moments = np.linalg.eigvalsh(nma.inertia_tensor)
        if self.symmetry_number == None:
            self.symmetry_number = nma.symmetry_number
            natom = len(nma.numbers)
            if self.symmetry_number == None and natom < 10:
                from molmod import Molecule
                # compute the rotational symmetry number
                tmp_mol = Molecule(nma.numbers, nma.coordinates)
                self.symmetry_number = tmp_mol.compute_rotsym()
            else:
                self.symmetry_number = 1
                print 'WARNING: molecule is too large (%i atoms > 10) to quickly estimate the rotational symmetry number.' % natom
        self.factor = np.sqrt(np.product([
            2*np.pi*m*boltzmann for m in self.moments if m > self.im_threshold
        ]))/self.symmetry_number/np.pi
        self.count = (self.moments > self.im_threshold).sum()

    def dump(self, f):
        """See :meth:`Info.dump`."""
        Info.dump(self, f)
        print >> f, "    Rotational symmetry number: %i" % self.symmetry_number
        print >> f, "    Moments of inertia [amu*bohr**2]: %f  %f %f" % tuple(self.moments/amu)
        print >> f, "    Threshold for non-zero moments of inertia [amu*bohr**2]: %e" % (self.im_threshold/amu)
        print >> f, "    Non-zero moments of inertia: %i" % self.count

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        if temp == 0:
            if n > 0:
                return 0.0
            else:
                raise NotImplementedError
        else:
            return float(temp)**n*(np.log(temp)*0.5*self.count + np.log(self.factor))

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        return float(temp)**(n-1)*0.5*self.count

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        return -float(temp)**(n-2)*0.5*self.count


class PCMCorrection(Info, StatFys):
    """A correction to the free energy as function of the temperature.

       The correction can be a constant shift of the free energy or a linear
       shift of the free energy as function of the temperature.
    """

    def __init__(self, point1, point2=None):
        """
           Argument:
            | ``point1`` -- A 2-tuple with free energy and a temperature. A
                            correction for the free energy at the given
                            temperature. (If no second point is given, the same
                            correction is applied to all temperatures.)

           Optional argument:
            | ``point2`` -- A 2-tuple with free energy and a temperature. In
                            combination with point1, a linear free energy
                            correction as function of the temperature is added.
        """
        if (not hasattr(point1, "__len__")) or len(point1) != 2:
            raise ValueError("The first argument must be a (delta_G, temp) pair.")
        if point2 is not None and ((not hasattr(point2, "__len__")) or len(point2)) != 2:
            raise ValueError("The second argument must be None or a (delta_G, temp) pair.")
        self.point1 = point1
        self.point2 = point2
        Info.__init__(self, "pcm_correction")

    def dump(self, f):
        """See :meth:`Info.dump`."""
        Info.dump(self, f)
        print >> f, "    Point 1:"
        print >> f, "       Delta G [kJ/mol]: %.2f" % (self.point1[0]/kjmol)
        print >> f, "       Temperature [K]: %.2f" % (self.point1[1])
        print >> f, "    Point 2:"
        if self.point2 is not None:
            print >> f, "       Delta G [kJ/mol]: %.2f" % (self.point2[0]/kjmol)
            print >> f, "       Temperature [K]: %.2f" % (self.point2[1])
        else:
            print >> f, "       Not Defined!! Only rely on computations on temperature of point 1!!"
        print >> f, "    Zero-point contribution [kJ/mol]: %.7f" % (self.zero_point_energy()/kjmol)

    def _eval_free(self, temp):
        if self.point2 is None:
            return self.point1[0], 0.0, 0.0
        else:
            slope = (self.point2[0]-self.point1[0])/(self.point2[1]-self.point1[1])
            return (
                self.point1[0] + slope*(temp-self.point1[1]),
                slope,
                0.0
            )

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        F, Fp, Fpp = self._eval_free(temp)
        return -F*float(temp)**(n-1)/boltzmann

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        F, Fp, Fpp = self._eval_free(temp)
        return (F*float(temp)**(n-2) - Fp*float(temp)**(n-1))/boltzmann

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        F, Fp, Fpp = self._eval_free(temp)
        return (-Fpp*float(temp)**(n-1) + 2*(Fp*float(temp)**(n-2) - F*float(temp)**(n-3)))/boltzmann


def helper_vibrations(temp, n, freqs, classical=False, freq_scaling=1, zp_scaling=1):
    """Helper 0 function for a set of harmonic oscillators.

       Returns T^n ln(Z), where Z is the partition function.

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``freqs`` -- an array with frequencies

       Optional arguments:
        | ``classical`` -- When True, the classical partition function is used.
                           [default=False]
        | ``freq_scaling`` -- Scale the frequencies with the given factor.
                              [default=1]
        | ``freq_zp`` -- Scale the zero-point energy correction with the given.
                         factor [default=1]
    """
    # this is defined as a function because multiple classes need it
    if classical:
        if temp == 0:
            if n >= 1:
                return np.zeros(len(freqs))
            else:
                raise NotImplementedError
        else:
            Af = planck*freqs*freq_scaling/(boltzmann*temp)
            return -float(temp)**n*np.log(Af)
    else:
        # The zero point correction is included in the vibrational partition
        # function.
        if temp == 0:
            Abis = freqs*(0.5*planck*zp_scaling/boltzmann)
            if n >= 1:
                return -Abis*float(temp)**(n-1)
            else:
                raise NotImplementedError
        else:
            A = freqs*(planck/(boltzmann*temp))
            B = np.exp(-freq_scaling*A)
            return -((0.5*zp_scaling)*A + np.log(1 - B))*float(temp)**n

def helpert_vibrations(temp, n, freqs, classical=False, freq_scaling=1, zp_scaling=1):
    """Helper 1 function for a set of harmonic oscillators.

       Returns T^n (d ln(Z) / dT), where Z is the partition function.

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``energy_levels`` -- an array with energy levels

       Optional arguments:
        | ``classical`` -- When True, the classical partition function is used
                           [default=False]
        | ``freq_scaling`` -- Scale the frequencies with the given factor
                              [default=1]
        | ``freq_zp`` -- Scale the zero-point energy correction with the given
                         factor [default=1]
    """
    # this is defined as a function because multiple classes need it
    if classical:
        if temp == 0:
            raise NotImplementedError
        else:
            result = float(temp)**(n-1)
            if hasattr(freqs, "__len__"):
                result *= np.ones(len(freqs))
            return result
    else:
        if temp == 0:
            raise NotImplementedError
        else:
            A = freqs*(planck/(boltzmann*temp))
            B = np.exp(-freq_scaling*A)
            C = B/(1 - B)
            return A*float(temp)**(n-1)*(0.5*zp_scaling + freq_scaling*C)

def helpertt_vibrations(temp, n, freqs, classical=False, freq_scaling=1, zp_scaling=1):
    """Helper 2 function for a set of harmonic oscillators.

       Returns T^n (d^2 ln(Z) / dT^2), where Z is the partition function.

       Arguments:
        | ``temp`` -- the temperature
        | ``n`` -- the power for the temperature factor
        | ``energy_levels`` -- an array with energy levels

       Optional arguments:
        | ``classical`` -- When True, the classical partition function is used
                           [default=False]
        | ``freq_scaling`` -- Scale the frequencies with the given factor
                              [default=1]
        | ``freq_zp`` -- Scale the zero-point energy correction with the given
                         factor [default=1]
    """
    # this is defined as a function because multiple classes need it
    if classical:
        if temp == 0:
            raise NotImplementedError
        else:
            result = -float(temp)**(n-2)
            if hasattr(freqs, "__len__"):
                result *= np.ones(len(freqs))
            return result
    else:
        if temp == 0:
            raise NotImplementedError
        else:
            A = freqs*(planck/(boltzmann*temp))
            Af = freq_scaling*A
            B = np.exp(-Af)
            C = B/(1.0 - B)
            return -A*float(temp)**(n-2)*(zp_scaling + freq_scaling*C*(2 - Af/(1-B)))


class Vibrations(Info, StatFysTerms):
    """The vibrational contribution to the partition function."""
    def __init__(self, classical=False, freq_scaling=1, zp_scaling=1, freq_threshold=None):
        """
           Optional arguments:
            | ``classical`` -- When True, the vibrations are treated classically
                               [default=False]
            | ``freq_scaling`` -- Scale factor for the frequencies [default=1]
            | ``zp_scaling`` -- Scale factor for the zero-point energy
                                correction [default=1]
            | ``freq_threshold`` -- Frequencies whose absolute value is below
                                    this threshold will be left out of the
                                    partition function, in addition to those
                                    already indicated as 'almost' zero by the
                                    NMA. This option is only needed to fix some
                                    pathological cases.
        """
        self.classical = classical
        self.freq_scaling = freq_scaling
        self.zp_scaling = zp_scaling
        self.freq_threshold = freq_threshold
        Info.__init__(self, "vibrational")

    def init_part_fun(self, nma, partf):
        """See :meth:`StatFys.init_part_fun`."""
        zero_indexes = nma.zeros
        nonzero_mask = np.ones(len(nma.freqs), dtype=bool)
        nonzero_mask[zero_indexes] = False
        if self.freq_threshold is not None:
            nonzero_mask[abs(nma.freqs) < self.freq_threshold] = False

        self.freqs = nma.freqs[nonzero_mask]
        self.zero_freqs = nma.freqs[~nonzero_mask]
        self.positive_freqs = nma.freqs[(nma.freqs > 0) & nonzero_mask]
        self.negative_freqs = nma.freqs[(nma.freqs < 0) & nonzero_mask]

        StatFysTerms.__init__(self, len(self.positive_freqs))

    def dump(self, f):
        """See :meth:`Info.dump`."""
        Info.dump(self, f)
        print >> f, "    Number of zero wavenumbers: %i " % (len(self.zero_freqs))
        print >> f, "    Number of real wavenumbers: %i " % (len(self.positive_freqs))
        print >> f, "    Number of imaginary wavenumbers: %i" % (len(self.negative_freqs))
        print >> f, "    Frequency scaling factor: %.4f" % self.freq_scaling
        print >> f, "    Zero-point scaling factor: %.4f" % self.zp_scaling
        self.dump_values(f, "Zero Wavenumbers [1/cm]", self.zero_freqs/(lightspeed/centimeter), "% 8.1f", 8)
        self.dump_values(f, "Real Wavenumbers [1/cm]", self.positive_freqs/(lightspeed/centimeter), "% 8.1f", 8)
        self.dump_values(f, "Imaginary Wavenumbers [1/cm]", self.negative_freqs/(lightspeed/centimeter), "% 8.1f", 8)
        print >> f, "    Zero-point contribution [kJ/mol]: %.7f" % (self.zero_point_energy()/kjmol)

    def helper_terms(self, temp, n):
        """See :meth:`StatFysTerms.helper_terms`."""
        return helper_vibrations(
            temp, n, self.positive_freqs, self.classical, self.freq_scaling,
            self.zp_scaling
        )

    def helpert_terms(self, temp, n):
        """See :meth:`StatFysTerms.helpert_terms`."""
        return helpert_vibrations(
            temp, n, self.positive_freqs, self.classical, self.freq_scaling,
            self.zp_scaling
        )

    def helpertt_terms(self, temp, n):
        """See :meth:`StatFysTerms.helpertt_terms`."""
        return helpertt_vibrations(
            temp, n, self.positive_freqs, self.classical, self.freq_scaling,
            self.zp_scaling
        )


class PartFun(Info, StatFys):
    """The partition function.

       This object contains all contributions to the partition function in
       self.terms and makes sure they are properly initialized. It also
       implements all the methods defined in StatFys, e.g. it can compute
       the entropy, the free energy and so on.
    """
    __reserved_names__ = set(["terms"])

    def __init__(self, nma, terms=None):
        """
           Arguments:
            | ``nma`` -- NMA object
           Optional arguments:
            | ``terms`` -- list to select the contributions to the partition
                           function e.g. [Vibrations(classical=True), ExtRot(1)]
        """
        if terms is None:
            terms = []
        self.terms = terms
        # perform a sanity check on the names of the contributions:
        for term in self.terms:
            if term.name in self.__reserved_names__:
                raise ValueError("An additional partition function term can not have the name '%s'" % term.name)
        # done testing, start initialization
        self.vibrational = None
        self.electronic = None
        for term in self.terms:
            self.__dict__[term.name] = term

        if self.vibrational is None:
            self.vibrational = Vibrations()
            self.terms.append(self.vibrational)

        if self.electronic is None:
            self.electronic = Electronic()
            self.terms.append(self.electronic)

        self.terms.sort(key=(lambda t: t.name))

        for term in self.terms:
            try:
                term.init_part_fun(nma, self)
            except Exception as err:
                if not err.args:
                   err.args=('',)
                err.args = ('%s: %s' % (term.name, err.message),)+err.args[1:]
                raise

        self.title = nma.title
        self.chemical_formula = nma.chemical_formula
        Info.__init__(self, "total")

    def helper(self, temp, n):
        """See :meth:`StatFys.helper`."""
        return sum(term.helper(temp, n) for term in self.terms)

    def helpert(self, temp, n):
        """See :meth:`StatFys.helpert`."""
        return sum(term.helpert(temp, n) for term in self.terms)

    def helpertt(self, temp, n):
        """See :meth:`StatFys.helpertt`."""
        return sum(term.helpertt(temp, n) for term in self.terms)

    def helpern(self, temp, n):
        """See :meth:`StatFys.helpern`."""
        return sum(term.helpern(temp, n) for term in self.terms)

    def helperv(self, temp, n):
        """See :meth:`StatFys.helperv`."""
        return sum(term.helperv(temp, n) for term in self.terms)

    def dump(self, f):
        """See :meth:`Info.dump`."""
        print >> f, "Title:", self.title
        print >> f, "Chemical formula:", self.chemical_formula
        print >> f, "Electronic energy [au]: %.5f" % self.electronic.energy
        print >> f, "Zero-point contribution [kJ/mol]: %.7f" % ((self.zero_point_energy() - self.electronic.energy)/kjmol)
        print >> f, "Zero-point energy [au]: %.5f" % self.zero_point_energy()
        print >> f, "Contributions to the partition function:"
        for term in self.terms:
            term.dump(f)

    def write_to_file(self, filename):
        """Write an extensive description of the parition function to a file.

           Argument:
            | ``filename`` -- The name of the file to write to.
        """
        f = file(filename, 'w')
        self.dump(f)
        f.close()



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--

from tamkin import *


def do_NMA(filecor, filehess, filechk, jobtimer):
    # Load the data.
    jobtimer.sample("start loading")
    jobtimer.dump()
    molecule = load_molecule_charmm(filecor,filehess)

    # Perform the normal mode analysis
    jobtimer.sample("nma")
    jobtimer.dump()
    nma = NMA(molecule)

    # Write to file
    jobtimer.sample("start writing")
    jobtimer.dump()
    nma.write_to_file(filechk)

    jobtimer.sample("done")
    jobtimer.dump()


def do_MBH(filecor, filehess, filefixed, filechk, jobtimer):
    # Load the data.
    jobtimer.sample("start loading")
    jobtimer.dump()
    molecule = load_molecule_charmm(filecor,filehess)
    jobtimer.sample("start loading fixed")
    fixed = load_blocks_txt(filefixed)

    # Perform the normal mode analysis
    jobtimer.sample("mbh")
    jobtimer.dump()
    nma = NMA(molecule, MBH(fixed))

    # Write to file
    jobtimer.sample("start writing")
    jobtimer.dump()
    nma.write_to_file(filechk)

    jobtimer.sample("done")
    jobtimer.dump()

def do_VSA(filecor, filehess, filefixed, filechk, jobtimer):
    # Load the data.
    jobtimer.sample("start loading")
    jobtimer.dump()
    molecule = load_molecule_charmm(filecor,filehess)
    jobtimer.sample("start loading fixed")
    subs = load_subs_txt(filefixed)

    # Perform the normal mode analysis
    jobtimer.sample("vsa")
    jobtimer.dump()
    nma = NMA(molecule, VSA(subs))

    # Write to file
    jobtimer.sample("start writing")
    jobtimer.dump()
    nma.write_to_file(filechk)

    jobtimer.sample("done")
    jobtimer.dump()

def take_cut(filechk,jobtimer):
    # Read the NMA object
    jobtimer.sample("start reading")
    jobtimer.dump()
    nma = NMA.read_from_file(filechk)

    jobtimer.sample("done")
    jobtimer.dump()


def do_gibbs(filechk, filethermo, jobtimer):
    jobtimer.sample("start reading")
    jobtimer.dump()
    nma = NMA.read_from_file(filechk)

    jobtimer.sample("gibbs")
    jobtimer.dump()
    pf = PartFun(nma, [Vibrations(), ExtTrans(), ExtRot(1)] )

    jobtimer.sample("thermo analysis")
    jobtimer.dump()
    ta = ThermoAnalysis(pf, [300.0])
    ta.write_to_file(filethermo)

    jobtimer.sample("done")
    jobtimer.dump()


def overlap(filecor1, filecor2, filechk, fileoverlap, jobtimer):

    # Get delta vector
    jobtimer.sample("delta")
    jobtimer.dump()
    delta = get_delta_vector_charmmcor(filecor1, filecor2, normalize=True)

    # Read the NMA object
    jobtimer.sample("start reading")
    jobtimer.dump()
    nma = NMA.read_from_file(filechk)

    # Calculate overlap
    jobtimer.sample("start overlap")
    jobtimer.dump()
    calculate_overlap(nma.modes,nma.freqs, delta,[0.0],filename=fileoverlap)

    jobtimer.sample("done")
    jobtimer.dump()



def get_delta(filecor1,filecor2,filedelta):
    jobtimer.sample("start delta")
    jobtimer.dump()
    get_delta_vector_charmmcor(filecor1, filecor2, normalize=True)



#-----------------------------------------
#------------ MAIN -----------------------

if __name__ == '__main__':
    from optparse import OptionParser

    usage = """
    python %prog FREQFILE [options]
    python %prog [options]

    This script ...
"""

    parser = OptionParser(usage=usage)

    parser.add_option('-o','--fig',dest='filename_fig',
                      default="intensity.png",
                      metavar="FIGNAME",
                      help="The name of the file that contains the plot, will be written.")

    parser.add_option('--job',dest='job',
                      default=None,
                      metavar="JOB",
                      help="Jobtype: nma, cut, ...")

    parser.add_option('--filecor', dest='filecor',
                      default = None,
                      metavar = "FILECOR",
                      help = "Name of file that contains the charmm coords.")

    parser.add_option('--filehess', dest='filehess',
                      default = None,
                      metavar = "FILEHESS",
                      help = "Name of file that contains the charmm Hessian.")

    parser.add_option('--filechk', dest='filechk',
                      default = None,
                      metavar = "FILECHK",
                      help = "Name of file that contains the checkpoint file made with Tamkin.")

    parser.add_option('--filecor2', dest='filecor2',
                      default = None,
                      metavar = "FILECOR2",
                      help = "Name of second file that contains the charmm coords.")

    parser.add_option('--filedelta', dest='filedelta',
                      default = None,
                      metavar = "FILEDELTA",
                      help = "Name of file that will contain the Delta (pacmanvector) between filecor and filecor2.")

    parser.add_option('--fileoverlap', dest='fileoverlap',
                      default = None,
                      metavar = "FILEOVERLAP",
                      help = "Name of file that will contain the overlap between the Delta (pacmanvector) and filecor (first cor file).")

    parser.add_option('--filefixed', dest='filefixed',
                      default = None,
                      metavar = "FILEFIXED",
                      help = "Name of file with partioning information eg blockchoice, environment atoms, ...")

    parser.add_option('--filethermo', dest='filethermo',
                      default = None,
                      metavar = "FILETHERMO",
                      help = "Name of file that will contain the thermodynamical analysis.")



    options, args = parser.parse_args()

    jobtimer = Timer()
    jobtimer.sample("start")



    if options.job == "nma":
        if None in [options.filecor, options.filehess, options.filechk]:
            print "at least one of the variables filecor, filehess or filechk is empty"
            print "Doing nothing..."
        else:
            do_NMA(options.filecor, options.filehess, options.filechk, jobtimer)

    if options.job == "mbh":
        if None in [options.filecor, options.filehess, options.filefixed, options.filechk]:
            print "at least one of the variables filecor, filehess, filefixed or filechk is empty"
            print "Doing nothing..."
        else:
            do_MBH(options.filecor, options.filehess, options.filefixed, options.filechk, jobtimer)

    if options.job == "vsa":
        if None in [options.filecor, options.filehess, options.filefixed, options.filechk]:
            print "at least one of the variables filecor, filehess, filefixed or filechk is empty"
            print "Doing nothing..."
        else:
            do_VSA(options.filecor, options.filehess, options.filefixed, options.filechk, jobtimer)

    elif options.job == "cut":
        if options.filechk is None:
            print "variable filechk is empty"
            print "Doing nothing..."
        else:
            take_cut(options.filechk, jobtimer)

    elif options.job == "delta":
        if None in [options.filecor, options.filecor2, options.filedelta]:
            print "at least one of the variables filecor, filecor2 or filedelta is empty"
            print "Doing nothing..."
        else:
            get_delta(options.filecor, options.filecor2, options.filedelta)


    elif options.job == "overlap":
        if None in [options.filecor, options.filecor2, options.filechk, options.fileoverlap]:
            print "at least one of the variables filecor, filecor2, filechk or fileoverlap is empty"
            print "Doing nothing..."
        else:
            overlap(options.filecor, options.filecor2,
                    options.filechk, options.fileoverlap, jobtimer)

    elif options.job == "gibbs":
        if None in [options.filechk, options.filethermo]:
            print "at least one of the variables filechk or filethermo is empty"
            print "Doing nothing..."
        else:
            do_gibbs(options.filechk, options.filethermo, jobtimer)

    elif options.job is None:
        print "no job specified"
        print "Doing nothing..."
    else:
        print "The job "+options.job+" is not defined/implemented. Doing nothing..."



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from molmod.test.test_examples import check_example


def test_example_001():
    check_example(__name__, "001_ethane", "thermo.py", ['gaussian.fchk'])

def test_example_002():
    check_example(__name__, "002_linear_co2", "thermo.py", ['gaussian.fchk'])

def test_example_003():
    check_example(__name__, "003_pentane", "thermo.py", ['gaussian.fchk'])

def test_example_005():
    check_example(__name__, "005_acrylamide_reaction", "reaction.py",
                  ['aa.fchk', 'aarad.fchk', 'paats.fchk'])

def test_example_006():
    check_example(__name__, "006_5T_ethyl_ethene_addition", "reaction.py",
                  ['react.fchk', 'ts.fchk'])

def test_example_007():
    check_example(__name__, "007_mfi_propene_reaction", "reaction.py",
                  ['Zp_p_react.28aug.com', 'Zp_p_TS.28aug.com', 'Zp_p_react.28aug.fchk',
                   'Zp_p_react.14mei.fchk', 'Zp_p_TS.28aug.fchk',
                   '5Tp_p_TS.oniom21apr_HF.fchk'])

def test_example_008():
    check_example(__name__, "008_ethane_rotor", "thermo.py",
                  ['freq/gaussian.fchk', 'scan/gaussian.log'])

def test_example_009():
    check_example(__name__, "009_ethyl_ethene", "reaction.py",
                  ['ethyl/freq/gaussian.fchk', 'ethene/freq/gaussian.fchk',
                   'ts_ad1/freq_gauche/gaussian.fchk', 'ts_ad1/freq_trans/gaussian.fchk',
                   'ethyl/scan/gaussian.log', 'ts_ad1/scan1/gaussian.log',
                   'ts_ad1/scan2/gaussian.log'])

def test_example_012():
    check_example(__name__, "012_ethyl_ethene_scaling", "reaction.py",
                  ['ethyl/freq/gaussian.fchk', 'ethene/freq/gaussian.fchk',
                   'ts_ad1/freq_gauche/gaussian.fchk'])

def test_example_013():
    check_example(__name__, "013_butane", "thermo.py",
                  ['freq/gaussian.fchk', 'scan/gaussian.log'])

def test_example_014():
    check_example(__name__, "014_pentane_mbh", "thermo.py", ['gaussian.fchk'])

def test_example_015():
    check_example(__name__, "015_kie", "reaction.py", ['reactant.fchk', 'trans.fchk'])

def test_example_016():
    check_example(__name__, "016_modes", "modes.py", ['PUNCH'])

def test_example_017():
    check_example(__name__, "017_activationkineticmodel", "reaction.py",
                  ['water/gaussian.fchk', 'VO_AA_OH_H2O/gaussian.fchk',
                   'VO_AA_OOtBu/gaussian.fchk', 'TBHP/gaussian.fchk', 'TS/TS.fchk',
                   'cyclohexene/gaussian.fchk'])

def test_example_018():
    check_example(__name__, "018_physisorption", "adsorption.py",
                  ['m062x/argon/gaussian.fchk', 'm062x/benzene/freq/gaussian.fchk',
                   'm062x/complex/freq/gaussian.fchk'])

def test_example_019():
    check_example(__name__, "019_ethyl_ethene_simple", "kinetic.py",
                  ['ethyl.fchk', 'ethene.fchk', 'ts_trans.fchk'])

def test_example_020():
    check_example(__name__, "020_butane_conformers", "equilibrium.py",
                  ['trans.fchk', 'gauche.fchk'])

def test_example_021():
    check_example(__name__, "021_water_formation", "formation.py",
                  ['oxygen.fchk', 'hydrogen.fchk', 'water.fchk'])



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--

import numpy as np, sys

from molmod.units import bar, liter, meter, mol, second, kjmol
from molmod.constants import boltzmann

__all__ = [
    "load_summary", "get_error_color",
    "temps", "invtemps", "experimental_k", "experimental_lnk", "experimental_A",
    "experimental_Ea", "covariance_parameters", "experimental_parameters",
]


def load_summary(fn):
    """Load a kinetics summary file

       Arguments:
         fn  --  a filename

       Returns:
         ks  --  The rate coefficients at 300, 400, 500 amd 600 K
         (A, Ea)  --  The pre-exponential factor and the activation energy
         E0, E  --  The zero-point corrected barrier and the classical barrier
    """
    f = file(fn)
    result = tuple(float(word) for word in f.next().split())
    f.close()
    return result[:-4], result[-4:-2], result[-2:]


def get_error_color(ratio):
    x = abs(ratio)
    if x < 1:
        rgb = (x,1.0,0.0)
    elif x < 2:
        rgb = (1.0,2.0-x,0.0)
    elif x < 3:
        rgb = ((4-x)/2,0.0,(x-2.0)/2)
    else:
        rgb = (0.5,0.0,0.5)
    result = "#%s" % "".join(hex(int(c*255))[2:].rjust(2,"0") for c in rgb)
    return result


temps = np.array([300, 400, 500, 600], float)
#temps = np.arange(300.0, 601.0, 10.0)
invtemps = 1/temps
N = len(temps)

centimeter = meter*0.01
experimental_k = 7.19e-15 * (temps/298)**2.44*np.exp(-22.45*kjmol/(boltzmann*temps))
experimental_k *= (centimeter**3/second)
experimental_k /= (meter**3/mol/second)
experimental_lnk = np.log(experimental_k)
print "unit conversion: %e" % (7.19e-15*(centimeter**3/second)/(meter**3/mol/second))
print "experimental_k:"
print experimental_k
print "experimental_lnk:"
print experimental_lnk

# fit experimental A and Ea and also sensitivity to errors

design_matrix = np.array([np.ones(N), -invtemps/boltzmann*kjmol]).transpose()
expected_values = experimental_lnk

A = np.dot(design_matrix.transpose(), design_matrix)
B = np.dot(design_matrix.transpose(), expected_values)

experimental_parameters = np.linalg.solve(A, B)
experimental_A = np.exp(experimental_parameters[0])
experimental_Ea = experimental_parameters[1]
covariance_lnk = np.ones((N,N),float)*np.log(10)**2 + np.identity(N,float)*np.log(2)**2
#covariance_lnk = np.identity(4,float)*np.log(10)**2
sensitivity = np.linalg.solve(A, design_matrix.transpose())
covariance_parameters = np.dot(np.dot(sensitivity, covariance_lnk), sensitivity.transpose())

print "experimental_A: %e" % experimental_A
print "experimental_Ea", experimental_Ea
#sys.exit()
# done fitting



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from tamkin.nma import NMA

from molmod import angstrom, lightspeed, centimeter
from molmod.io import XYZWriter
from molmod.periodic import periodic

import numpy as np


__all__ = ["dump_modes_xyz", "dump_modes_molden", "dump_modes_gaussian"]


def dump_modes_xyz(nma, indexes=0, prefix="mode", amplitude=5.0*angstrom, frames=36):
    """Write XYZ trajectory file(s) that vizualize internal mode(s)

       Arguments:
         | nma  --  an object that specifies the normal modes, several formats
                    are supported: (i) a Tamkin NMA object, (ii) a 3-tuple with
                    reference coordinates, mass-unweighted mode(s) and atom
                    numbers or (iii) a 4-tuple with reference coordinates, mass-
                    weighted mode(s), atom numbers and a masses3 vector. the
                    latter is a vector with 3*N elements containing the masses of
                    the atoms in groups of three.

       Optional arguments:
         | indexes  --  the index or a list of indexes of modes that must be
                        written to trajectory files [default=0]
         | prefix  --  a prefix used for the output files. the generated
                       trajectory filenames have the format prefix.index.xyz
                       [default="mode"]
         | amplitude  --  the amplitude of the normal mode vibration in atomic
                          untis [default=5*angstrom]
         | frames  --  the number of frames written to the trajectory file
                       [default=36]
    """

    if isinstance(nma, NMA):
        coordinates = nma.coordinates
        modes = nma.modes
        numbers = nma.numbers
        masses3 = nma.masses3
    elif hasattr(nma, "__len__") and len(nma)==3:
        coordinates, modes, numbers = nma
        masses3 = None
    elif hasattr(nma, "__len__") and len(nma)==4:
        coordinates, modes, numbers, masses3 = nma
    else:
        raise TypeError("Could not understand first argument. Check documentation.")

    if not hasattr(indexes, "__len__"):
        indexes = [indexes]

    if len(modes.shape) == 1:
        modes = modes.reshape((-1,1))

    symbols = [periodic[n].symbol for n in numbers]

    for index in indexes:
        filename = "%s.%i.xyz" % (prefix, index)
        mode = modes[:,index]
        if masses3 is not None:
            mode /= np.sqrt(masses3)
        mode /= np.linalg.norm(mode)
        xyz_writer = XYZWriter(filename, symbols)
        for frame in xrange(frames):
            factor = amplitude*np.sin(2*np.pi*float(frame)/frames)
            xyz_writer.dump("frame %i" % frame, coordinates + factor*mode.reshape((-1,3)))
        del xyz_writer


def dump_modes_gaussian(filename, nma, selected=None):
    """Write freqs and modes to a file in the Gaussian log format.

       Arguments:
         | filename  --  modes are written to this file,
                         can be read by Molden (visualization program)
         | nma  --  an NMA object or a tuple or list with five elements: modes,
                    frequencies, masses, numbers, coordinates. See
                    _make_moldenfile for details.

       Optional argument:
         | selected  --  Selection of modes for which to make
                         trajectories. This can be a list or array
                         of mode indices (length <= N), or an
                         array of booleans (length = N).

       The output file will look like a stripped Gaussian03 or Gaussian09 log
       file. It is sufficient to visualize the modes in Molden, Openbabel or
       Avogadro.
    """
    ### A) Parse the NMA argument
    if isinstance(nma, NMA):
        # NMA object
        modes, freqs, masses, numbers, coordinates = \
            nma.modes, nma.freqs, nma.masses, nma.numbers, nma.coordinates
    elif hasattr(nma, "__len__") and len(nma) == 5 and not isinstance(nma, np.ndarray):
        # [modes, freqs, ...] or (modes, freqs, ...)
        modes, freqs, masses, numbers, coordinates = nma
    else:
        raise TypeError("nma argument has wrong type")

    ### B) Select some modes
    if selected is not None:
        modes = np.take(modes, selected, 1)  # modes in columns
        freqs = freqs[selected]
        masses = masses[selected]
        numbers = numbers[selected]
        coordinates = coordinates[selected]

    ### C) convert modes to the right convention
    masses3_sqrt1 = np.array(sum([[1/m,1/m,1/m] for m in np.sqrt(masses)],[]))
    nmode = modes.shape[1]
    modes = modes.copy() # avoid modifying the given modes
    for imode in xrange(nmode):
        modes[:,imode] *= masses3_sqrt1
        modes[:,imode] /= np.linalg.norm(modes[:,imode])

    ### D) Define some multiline text blobs that are used below.
    header = """\
 Entering Gaussian System, Link 0=g09

 This file is generated from the dump_modes_gaussian function in the file
 trajectory.py of TAMkin. Please note, that this is a "fake" output; TAMkin
 doesn't compute intensities. This file should be readable by Molden, Openbabel
 and Avogadro.

 Gaussian, Inc

 # Fake method line for openbabel
 """

    header_coordinates = """\
                         Standard orientation:
 ---------------------------------------------------------------------
 Center     Atomic      Atomic             Coordinates (Angstroms)
 Number     Number       Type             X           Y           Z
 ---------------------------------------------------------------------"""

    header_basisfunctions = """\
 ---------------------------------------------------------------------
       basis functions          primitive gaussians
       alpha electrons          beta electrons
 **********************************************************************"""

    header_freq = """\
 Harmonic frequencies (cm**-1), IR intensities (KM/Mole),
 Raman scattering activities (A**4/AMU), Raman depolarization ratios,
 reduced masses (AMU), force constants (mDyne/A) and normal coordinates:"""

    txt_freqbelow_1 = """\
 Red. masses --     0.0000
 Frc consts  --     0.0000
 IR Inten    --     0.0000
  Atom  AN      X      Y      Z"""
    txt_freqbelow_2 = """\
 Red. masses --     0.0000                 0.0000
 Frc consts  --     0.0000                 0.0000
 IR Inten    --     0.0000                 0.0000
  Atom  AN      X      Y      Z        X      Y      Z"""
    txt_freqbelow_3 = """\
 Red. masses --     0.0000                 0.0000                 0.0000
 Frc consts  --     0.0000                 0.0000                 0.0000
 IR Inten    --     0.0000                 0.0000                 0.0000
  Atom  AN      X      Y      Z        X      Y      Z        X      Y      Z"""
    txt_freqbelow = [txt_freqbelow_1, txt_freqbelow_2, txt_freqbelow_3]

    ### E) Actual writing of the fake log file
    with open(filename, "w") as f:
        print >> f, header

        ### E1) ATOM PART
        assert modes.shape[0] % 3 == 0
        natom = modes.shape[0]/3
        print >> f, header_coordinates
        for iatom in range(natom):
           print >> f, '%5d %10d %13s %15f %11f %11f' %(
                       iatom + 1, numbers[iatom], "0",
                       coordinates[iatom,0]/angstrom,
                       coordinates[iatom,1]/angstrom,
                       coordinates[iatom,2]/angstrom)

        ### E2) ORBITAL PART
        print >> f, header_basisfunctions
        print >> f, " "   #this part is just empty

        ### D3) FREQUENCY PART
        # Modes are written in sections with three columns each.
        print >> f, header_freq
        # istart: the current mode index, will be incremented by 3 in each iteration.
        istart = 0
        while istart < nmode:
            # select modes and freqs for this block
            iend = min(istart+3, nmode)
            ncol = iend - istart # number of columns in section
            # print stuff to file
            #  - mode indexes
            for imode in xrange(istart, iend):
                print >> f, '%22d' % (imode + 1),
            print >> f
            #  - (fake) symmetry info
            print >> f, ' '.join(["                    ?A"]*ncol)
            #  - frequencies converted to inverse centimeters
            print >> f, ' Frequencies --',
            for imode in xrange(istart, iend):
                print >> f, '%10.4f' % (freqs[imode]/lightspeed*centimeter),
                if imode != iend-1:
                    print >> f, '           ',
            print >> f
            #  - a blob of text below the frequencies
            print >> f, txt_freqbelow[ncol-1]
            #  - the modes
            for iatom in range(natom):
                print >> f, '%6d %3d' % (iatom + 1, numbers[iatom]),
                for imode in xrange(istart, iend):
                    print >> f, '%8.2f %6.2f %6.2f' % (
                       modes[3*iatom, imode], modes[3*iatom+1, imode],
                       modes[3*iatom+2, imode]),
                print >> f
            # prepare for next iteration
            istart = iend

        print >> f, " Normal termination of Gaussian 09."

# For backward compatibility with TAMkin before version 1.0.2
dump_modes_molden = dump_modes_gaussian
